#include "ssh_inc.hpp"
#include "libsftp.hpp"

#include <sys/wait.h>
#include <errno.h>
#include <iostream>

extern "C" {
  int showprogress = 0;
  int interrupted = 0;
}

namespace {
  void
  START_MARKER(const char *name)
  {
#if 0
    std::cerr << "[BEGIN] " << name << "\n";
#endif
  }

  int
  RETURN(int error, const char *name)
  {
#if 0
    std::cerr << "[RETURN] [error = " << error << "] " << name << "\n";
#endif
    return error;
  }

  void
  convRawAttrib2SFTPAttrib(SFTP::Attrib *dest, const ::Attrib &src)
  {
    memset(dest, 0, sizeof(SFTP::Attrib));

    if (src.flags & SSH2_FILEXFER_ATTR_SIZE) {
      dest->size = src.size;
    }
    if (src.flags & SSH2_FILEXFER_ATTR_UIDGID) {
      dest->uid = src.uid;
      dest->gid = src.gid;
    }
    if (src.flags & SSH2_FILEXFER_ATTR_PERMISSIONS) {
      dest->perm = src.perm;
    }
    if (src.flags & SSH2_FILEXFER_ATTR_ACMODTIME) {
      dest->atime = src.atime;
      dest->mtime = src.mtime;
    }
  }

  void
  convSFTPAttrib2RawAttrib(::Attrib *dest, const SFTP::Attrib &src)
  {
    memset(dest, 0, sizeof(::Attrib));

    dest->flags |= SSH2_FILEXFER_ATTR_SIZE;
    dest->size = src.size;

#if 0
    dest->flags |= SSH2_FILEXFER_ATTR_UIDGID;
    dest->uid = src.uid;
    dest->gid = src.gid;
#endif

    dest->flags |= SSH2_FILEXFER_ATTR_PERMISSIONS;
    dest->perm = src.perm;

    dest->flags |= SSH2_FILEXFER_ATTR_ACMODTIME;
    dest->atime = src.atime;
    dest->mtime = src.mtime;
  }

  char *
  cppstring2cstring(const std::string &str)
  {
    return const_cast<char *>(str.c_str());
  }
}

// ======================================================================
SFTP::Handler::Handler(void)
{
  conn = NULL;
}

bool
SFTP::Handler::initialize(const std::string &sshhostpath)
{
  const size_t copy_buffer_len = 32768;
  const size_t num_requests = 16;

  parseSSHHostPath(sshhostpath);

  int fd = makeSocket();
  if (fd == -1) {
    return false;
  }

  conn = do_init(fd, fd, copy_buffer_len, num_requests);
  if (conn == NULL) {
    return false;
  }

  int error = stat(NULL, ".");
  if (error) return false;

  return true;
}

void
SFTP::Handler::parseSSHHostPath(const std::string &sshhostpath)
{
  std::string::size_type colon = sshhostpath.find(':');

  if (colon == std::string::npos) {
    basepath = "./";
    hostname = sshhostpath;
  } else {
    basepath = sshhostpath.substr(colon + 1);
    hostname = sshhostpath.substr(0, colon);
  }

  if (basepath.size() > 0 && basepath[basepath.length() - 1] != '/') {
    basepath += '/';
  }

#ifdef DEBUG
  std::cerr << "hostname: " << hostname << "\n";
  std::cerr << "basepath: " << basepath << "\n";
#endif
}

int
SFTP::Handler::makeSocket(void)
{
  int sockpair[2];

  if (socketpair(AF_UNIX, SOCK_STREAM, 0, sockpair) == -1) {
    std::cerr << "[ERROR] socketpair failed" << "\n";
    return -1;
  }

  int pid = fork();
  if (pid == -1) {
    std::cerr << "[ERROR] fork failed" << "\n";
    return -1;

  } else if (pid == 0) {
    if (dup2(sockpair[1], 0) == -1 || dup2(sockpair[1], 1) == -1) {
      std::cerr << "[ERROR] failed to redirect input/output" << "\n";
      exit(1);
    }
    close(sockpair[0]);
    close(sockpair[1]);

#ifdef DEBUG
    int devnull = open("/dev/null", O_WRONLY);
    if (devnull != -1) {
      dup2(devnull, 2);
    }
    close(devnull);
#endif

    switch (fork()) {
    case -1:
      perror("failed to fork");
      exit(1);
    case 0:
      break;
    default:
      exit(0);
    }

    chdir("/");
    const char *argv[] = {
      "ssh",
      "-oServerAliveInterval=30",
      hostname.c_str(),
      "-s", "sftp",
      NULL,
    };
    execvp("ssh", const_cast<char * const *>(argv));
    perror("execvp");
    exit(0);
  }

  waitpid(pid, NULL, 0);
  close(sockpair[1]);

  return sockpair[0];
}

void
SFTP::Handler::terminate(void)
{
  xfree(conn);
}

void
SFTP::Handler::adjustRemotePath(std::string *adjusted, const std::string &remote)
{
  if (remote.size() == 0) {
    *adjusted = basepath;
  } else {
    if (remote[0] == '/') {
      *adjusted = remote;
    } else {
      *adjusted = basepath + remote;
    }
  }

#ifdef DEBUG
  std::cerr << "adjusted = " << *adjusted << "\n";
#endif
}

template <SFTP::Handler::statfunc_t _F> int
SFTP::Handler::internal_stat(Attrib *attr, const std::string &remote)
{
  std::string adjusted;
  adjustRemotePath(&adjusted, remote);

  const int quiet = 1;
  ::Attrib *a = _F(conn, cppstring2cstring(adjusted), quiet);
  if (a == NULL) return ENOENT;

  if (attr) {
    convRawAttrib2SFTPAttrib(attr, *a);
  }

  return 0;
}

// ======================================================================
int
SFTP::Handler::stat(Attrib *attr, const std::string &remote)
{
  START_MARKER(__PRETTY_FUNCTION__);

  if (! conn) return EIO;

  return internal_stat<do_stat>(attr, remote);
}

int
SFTP::Handler::lstat(Attrib *attr, const std::string &remote)
{
  START_MARKER(__PRETTY_FUNCTION__);

  if (! conn) return EIO;

  return internal_stat<do_lstat>(attr, remote);
}

int
SFTP::Handler::download(const std::string &remote, const std::string &local, bool pflag)
{
  START_MARKER(__PRETTY_FUNCTION__);

  if (! conn) return EIO;

  std::string adjusted;
  adjustRemotePath(&adjusted, remote);

  int error = do_download(conn, cppstring2cstring(adjusted), cppstring2cstring(local), pflag);
  if (error) return EIO;

  return 0;
}

int
SFTP::Handler::upload(const std::string &local, const std::string &remote, bool pflag)
{
  START_MARKER(__PRETTY_FUNCTION__);

  if (! conn) return EIO;

  std::string adjusted;
  adjustRemotePath(&adjusted, remote);

  int error = do_upload(conn, cppstring2cstring(local), cppstring2cstring(adjusted), pflag);
  if (error) return EIO;

  return 0;
}

int
SFTP::Handler::readdir(std::vector<DirentPtr> *dirent, const std::string &remote)
{
  START_MARKER(__PRETTY_FUNCTION__);

  if (! conn) return EIO;

  std::string adjusted;
  adjustRemotePath(&adjusted, remote);

  SFTP_DIRENT **dir;

  int error = do_readdir(conn, cppstring2cstring(adjusted), &dir);
  if (error) return EIO;

  dirent->clear();
  for (int i = 0; dir[i] != NULL; ++i) {
    DirentPtr p = DirentPtr(new Dirent);
    p->filename = dir[i]->filename;
    p->longname = dir[i]->longname;
    convRawAttrib2SFTPAttrib(&(p->a), dir[i]->a);
    dirent->push_back(p);
  }

  free_sftp_dirents(dir);

  return 0;
}

int
SFTP::Handler::readlink(std::string *remoteReferred, const std::string &remoteLink)
{
  START_MARKER(__PRETTY_FUNCTION__);

  if (! conn) return EIO;

  std::string adjusted;
  adjustRemotePath(&adjusted, remoteLink);
#ifdef DEBUG
  std::cerr << "readlink remoteLink = " << adjusted << "\n";
#endif

  char *referred = do_readlink(conn, cppstring2cstring(adjusted));
  if (! referred) {
    return EIO;
  }

#ifdef DEBUG
  std::cerr << "readlink referred = " << referred << "\n";
#endif

  *remoteReferred = referred;
  xfree(referred);

  return 0;
}

int
SFTP::Handler::symlink(const std::string &linktarget, const std::string &remote)
{
  START_MARKER(__PRETTY_FUNCTION__);

  if (! conn) return EIO;

  // checking file.
  int error = lstat(NULL, remote);
  if (! error) {
    return EEXIST;
  }

  std::string adjusted;
  adjustRemotePath(&adjusted, remote);

  error = do_symlink(conn, cppstring2cstring(linktarget), cppstring2cstring(adjusted));
  if (error) {
    return EIO;
  }

  return 0;
}


int
SFTP::Handler::rm(const std::string &remote)
{
  START_MARKER(__PRETTY_FUNCTION__);

  if (! conn) return EIO;

  std::string adjusted;
  adjustRemotePath(&adjusted, remote);

  u_int error = do_rm(conn, cppstring2cstring(adjusted));
  if (error) return EIO;

  return 0;
}

int
SFTP::Handler::mkdir(const std::string &remote)
{
  START_MARKER(__PRETTY_FUNCTION__);

  if (! conn) return EIO;

  int error = stat(NULL, remote);
  if (! error) {
    return EEXIST;
  }

  std::string adjusted;
  adjustRemotePath(&adjusted, remote);

  ::Attrib a;
  attrib_clear(&a);
  a.flags |= SSH2_FILEXFER_ATTR_PERMISSIONS;
  a.perm = 0777;

  error = do_mkdir(conn, cppstring2cstring(adjusted), &a);
  if (error) return EIO;

  return 0;
}

int
SFTP::Handler::rmdir(const std::string &remote)
{
  START_MARKER(__PRETTY_FUNCTION__);

  if (! conn) return EIO;

  std::string adjusted;
  adjustRemotePath(&adjusted, remote);

  int error = do_rmdir(conn, cppstring2cstring(adjusted));
  if (error) return EIO;

  return 0;
}

int
SFTP::Handler::rename(const std::string &remoteOld, const std::string &remoteNew)
{
  START_MARKER(__PRETTY_FUNCTION__);

  if (! conn) return EIO;

  std::string adjustedOld;
  adjustRemotePath(&adjustedOld, remoteOld);

  std::string adjustedNew;
  adjustRemotePath(&adjustedNew, remoteNew);

  // if remoteNew exists, then remote first.
  int error = stat(NULL, remoteNew);
  if (! error) {
    error = rm(remoteNew);
    if (error) return EIO;
  }

  error = do_rename(conn, cppstring2cstring(adjustedOld), cppstring2cstring(adjustedNew));
  if (error) return EIO;

  return 0;
}

int
SFTP::Handler::setattr(const std::string &remote, const Attrib &attr)
{
  START_MARKER(__PRETTY_FUNCTION__);

  if (! conn) return EIO;

  std::string adjusted;
  adjustRemotePath(&adjusted, remote);

  ::Attrib a;
  convSFTPAttrib2RawAttrib(&a, attr);
  int error = do_setstat(conn, cppstring2cstring(adjusted), &a);

  if (error) return EIO;

  return 0;
}
