#ifndef URDFS_BRIDGE_HPP
#define URDFS_BRIDGE_HPP

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <sys/vnode.h>

namespace URDFS {
  namespace Bridge {
    extern const char *fsname;
    enum {
      ROOT_INO = 2,
    };

    struct MountArgs
    {
      char *mntfromname;
      int version;
      int socketNameLen;
      struct sockaddr *socketName;
      char *volName;
      u_int32_t flags;
      off_t dirSize;
    };

    enum FileSystemOperation {
      URDFS_LOOKUP,
      URDFS_CREATE,
      URDFS_OPEN,
      URDFS_CLOSE,
      URDFS_GETATTR,
      URDFS_SETATTR,
      URDFS_WRITE,
      URDFS_FSYNC,
      URDFS_REMOVE,
      URDFS_RENAME,
      URDFS_MKDIR,
      URDFS_RMDIR,
      URDFS_READDIR,
      URDFS_READLINK,
      URDFS_SYMLINK,
      URDFS_STATFS,
      URDFS_UNMOUNT,
      URDFS_INVALCACHES,
      // for the future
      URDFS_LINK,
      URDFS_MKNOD,
      URDFS_GETATTRLIST,
      URDFS_SETATTRLIST,
      URDFS_EXCHANGE,
      URDFS_READDIRATTR,
      URDFS_SEARCHFS,
      URDFS_COPYFILE,
    };

    namespace Lookup {
      struct Request {
        ino_t parentIno;
        char name[NAME_MAX + 1];
      };

      struct Reply {
        ino_t ino;
        enum vtype vtype;
        struct timespec atime;
        struct timespec mtime;
        struct timespec ctime;
        off_t         filesize;
      };
    };

    namespace Create {
      struct Request {
        ino_t parentIno;
        mode_t mode;
        char name[NAME_MAX + 1];
      };

      struct Reply {
        ino_t ino;
      };
    }

    namespace Open {
      struct Request {
        ino_t ino;
        int flags;
      };

      struct Reply {
        char cacheFilePath[PATH_MAX];
      };
    }

    namespace Close {
      struct Request {
        ino_t ino;
      };

      struct Reply {};
    }

    namespace GetAttr {
      struct Request {
        ino_t ino;
      };

      struct Reply {
        struct stat attr;
      };
    }

    namespace SetAttr {
      struct Request {
        ino_t ino;

        bool isSetMode;
        mode_t mode;

        bool isSetSize;
        off_t size;

        bool isSetAtime;
        struct timespec atime;

        bool isSetMtime;
        struct timespec mtime;
      };

      struct Reply {};
    }

    namespace Fsync {
      struct Request {
        ino_t ino;
      };

      struct Reply {};
    }

    namespace Remove {
      struct Request {
        ino_t ino;
      };

      struct Reply {};
    }

    namespace Rename {
      struct Request {
        ino_t from_ino;
        ino_t from_parentIno;
        ino_t to_ino;
        ino_t to_parentIno;
        char to_name[NAME_MAX + 1];
      };

      struct Reply {};
    }

    namespace MkDir {
      struct Request {
        ino_t parentIno;
        mode_t mode;
        char name[NAME_MAX + 1];
      };

      struct Reply {
        ino_t ino;
      };
    }

    namespace RmDir {
      struct Request {
        ino_t ino;
      };

      struct Reply {};
    }

    namespace ReadDir {
      struct Request {
        ino_t ino;
      };

      struct Reply {};
    }

    namespace ReadLink {
      struct Request {
        ino_t ino;
      };

      struct Reply {
        char path[PATH_MAX];
      };
    }

    namespace SymLink {
      struct Request {
        ino_t parentIno;
        char name[NAME_MAX + 1];
        char linktarget[PATH_MAX];
      };

      struct Reply {
        ino_t ino;
      };
    }

    namespace StatFs {
      struct Request {};

      struct Reply {
        struct vfsstatfs fs_attr;
      };
    }

    namespace UnMount {
      struct Request {};
      struct Reply {};
    }

    namespace InvalCaches {
      struct Request {};
      struct Reply{};
    }

    union Request
    {
      Lookup::Request lookup;
      Create::Request create;
      Open::Request open;
      Close::Request close;
      GetAttr::Request getattr;
      SetAttr::Request setattr;
      Fsync::Request fsync;
      Remove::Request remove;
      Rename::Request rename;
      MkDir::Request mkdir;
      RmDir::Request rmdir;
      ReadDir::Request readdir;
      ReadLink::Request readlink;
      SymLink::Request symlink;
      StatFs::Request statfs;
      UnMount::Request unmount;
      InvalCaches::Request invalcaches;
    };
  }
}

#endif
