#include "urdfs_kext.hpp"
#include "urdfs_kext_bridge.hpp"

static void
releaseSocket(socket_t socket)
{
  sock_shutdown(socket, SHUT_RDWR);
  sock_close(socket);
}

static bool
makeSocket(socket_t *socket)
{
  int error = sock_socket(PF_LOCAL, SOCK_STREAM, 0, NULL, NULL, socket);
  if (error) {
    printf("URDFS::sendmsg: sock_socket() = %d\n", error);
    return false;
  }

  // ----------------------------------------
  struct timeval tv;
  tv.tv_sec = URDFS::Kernel::Bridge::BRIDGE_TIMEOUT;
  tv.tv_usec = 0;

  error = sock_setsockopt(*socket, SOL_SOCKET, SO_RCVTIMEO, &tv, sizeof(struct timeval));
  if (error) {
    releaseSocket(*socket);
    printf("URDFS::sendmsg: sock_setsockopt() = %d\n", error);
    return false;
  }

  return true;
}

static void
dead(URDFS::Kernel::MountInfo *mountinfo)
{
  if (mountinfo->aliveFlag) {
    printf("server: %s: connection is dead\n", vfs_statfs(mountinfo->mountp)->f_mntfromname);
    vfs_event_signal(&vfs_statfs(mountinfo->mountp)->f_fsid, VQ_DEAD, 0);
    mountinfo->aliveFlag = false;
  }
}

static bool
connectSocket(socket_t socket, URDFS::Kernel::MountInfo *mountinfo)
{
  int error = sock_connect(socket, mountinfo->socketName, 0);
  if (error && error != EINPROGRESS) {
    if (error == ECONNREFUSED) {
      dead(mountinfo);
    }
    if (error != ENOENT) {
      printf("URDFS::sendmsg: sock_connect() = %d\n", error);
    }
    return false;
  }
  error = sock_nointerrupt(socket, TRUE);
  if (error) {
    printf("URDFS::sendmsg: sock_nointerrupt() = %d\n", error);
    return false;
  }

  return true;
}

int
URDFS::Kernel::Bridge::sendmsg(int vnop, MountInfo *mountinfo,
                               void *request, size_t requestsize,
                               void *reply, size_t replysize)
{
  START_MARKER(__PRETTY_FUNCTION__);

  int result;

  if (mountinfo == NULL) {
    return RETURN(ENXIO, __PRETTY_FUNCTION__);
  }
  if (vnop != URDFS::Bridge::URDFS_UNMOUNT && vfs_isforce(mountinfo->mountp)) {
    return RETURN(ENXIO, __PRETTY_FUNCTION__);
  }

  socket_t socket;
  if (! makeSocket(&socket)) {
    return RETURN(EIO, __PRETTY_FUNCTION__);
  }

  if (! connectSocket(socket, mountinfo)) {
    releaseSocket(socket);
    return RETURN(EIO, __PRETTY_FUNCTION__);
  }

  struct msghdr msg;
  memset(&msg, 0, sizeof(msg));

  struct iovec aiov[3];
  size_t iolen;
  aiov[0].iov_base = reinterpret_cast<caddr_t>(&vnop);
  aiov[0].iov_len = sizeof(vnop);
  aiov[1].iov_base = reinterpret_cast<caddr_t>(&requestsize);
  aiov[1].iov_len = sizeof(requestsize);
  aiov[2].iov_base = reinterpret_cast<caddr_t>(request);
  aiov[2].iov_len = requestsize;
  msg.msg_iovlen = 3;
  msg.msg_iov = aiov;

  int error = sock_send(socket, &msg, 0, &iolen);
  if (error) {
    printf("URDFS::sendmsg: sock_send() = %d\n", error);
    goto finish;
  }

  memset(&msg, 0, sizeof(msg));

  aiov[0].iov_base = reinterpret_cast<caddr_t>(&result);
  aiov[0].iov_len = sizeof(result);
  aiov[1].iov_base = reinterpret_cast<caddr_t>(reply);
  aiov[1].iov_len = replysize;
  msg.msg_iov = aiov;
  msg.msg_iovlen = (replysize == 0 ? 1 : 2);

  for (;;) {
    error = sock_receive(socket, &msg, MSG_WAITALL, &iolen);

    if (error == EWOULDBLOCK) {
      continue;
    }

    break;
  }
  if (error) goto finish;

 finish:
  releaseSocket(socket);

  if (error) return RETURN(EIO, __PRETTY_FUNCTION__);
  return RETURN(result, __PRETTY_FUNCTION__);
}
