#include "urdfs_kext.hpp"

extern "C" {
#include <sys/proc.h>
#include <sys/malloc.h>
#include <sys/file.h>
#include <sys/sysctl.h>
#include <libkern/libkern.h>
  kern_return_t urdfs_module_start(struct kmod_info *ki, void *data);
  kern_return_t urdfs_module_stop(struct kmod_info *ki, void *data);
}

#include "urdfs_kext_node.hpp"
#include "urdfs_kext_bridge.hpp"

// ======================================================================
namespace URDFS {
  namespace Kernel {
    static long mountCount = 0;
    static vfstable_t vfsconf;

    extern struct vnodeopv_desc vnodeop_opv_desc;
    static struct vnodeopv_desc *vnodeop_opv_desc_list[1] = {
      &vnodeop_opv_desc
    };
  }
}

namespace {
  // --------------------------------------------------
  int
  urdfs_init(struct vfsconf *vfsp)
  {
#pragma unused(vfsp)
    START_MARKER(__PRETTY_FUNCTION__);

    URDFS::NodeHash::initialize();

    return RETURN(0, __PRETTY_FUNCTION__);
  }

  // --------------------------------------------------
  bool
  createRootVnode(struct mount *mp, URDFS::Kernel::MountInfo *mountinfo)
  {
    START_MARKER(__PRETTY_FUNCTION__);

    struct timeval tv;
    struct timespec ts;

    microtime(&tv);
    TIMEVAL_TO_TIMESPEC(&tv, &ts);

    vnode_t rootvp;
    int error = URDFS::Kernel::getVnode(mp, NULLVP, 1, NULL, URDFS::Bridge::ROOT_INO, VDIR, ts, ts, ts, mountinfo->dirSize, &rootvp);
    if (error) return false;

    error = vnode_ref(rootvp);
    vnode_put(rootvp);
    if (error) return false;

    mountinfo->rootVnode = rootvp;
    return true;
  }

  // --------------------------------------------------
  int
  copyInMountArgs(URDFS::Kernel::MountArgs *args, user_addr_t data, vfs_context_t context)
  {
    START_MARKER(__PRETTY_FUNCTION__);

    if (vfs_context_is64bit(context)) {
      int error = copyin(data, args, sizeof(*args));

      return RETURN(error, __PRETTY_FUNCTION__);

    } else {
      URDFS::Bridge::MountArgs args_32;
      int error = copyin(data, (caddr_t)&args_32, sizeof(args_32));
      if (error) return RETURN(error, __PRETTY_FUNCTION__);

      args->mntfromname         = CAST_USER_ADDR_T(args_32.mntfromname);
      args->version             = args_32.version;
      args->socketNameLen       = args_32.socketNameLen;
      args->socketName          = CAST_USER_ADDR_T(args_32.socketName);
      args->volName             = CAST_USER_ADDR_T(args_32.volName);
      args->flags               = args_32.flags;
      args->dirSize             = args_32.dirSize;

      return RETURN(0, __PRETTY_FUNCTION__);
    }
  }

  int
  urdfs_mount(struct mount *mp, vnode_t devvp, user_addr_t data, vfs_context_t context)
  {
#pragma unused(devvp)
    START_MARKER(__PRETTY_FUNCTION__);

    ++URDFS::Kernel::mountCount;

    int error = 0;
    URDFS::Kernel::MountArgs args;
    URDFS::Kernel::MountInfo *mountinfo = NULL;
    size_t size;

    if (vfs_isupdate(mp)) {
      error = ENOTSUP;
      goto bad;
    }

    // ----------------------------------------
    error = copyInMountArgs(&args, data, context);
    if (error) goto bad;

    if (args.version != 1) {
      error = EINVAL;
      goto bad;
    }

    MALLOC(mountinfo, URDFS::Kernel::MountInfo *, sizeof(URDFS::Kernel::MountInfo), M_TEMP, M_WAITOK);
    memset(mountinfo, 0, sizeof(URDFS::Kernel::MountInfo));

    mountinfo->aliveFlag = true;
    mountinfo->mountp = mp;

    // Get the server sockaddr from the args
    MALLOC(mountinfo->socketName, struct sockaddr *, args.socketNameLen, M_TEMP, M_WAITOK);
    error = copyin(args.socketName, mountinfo->socketName, args.socketNameLen);
    if (error) {
      goto bad;
    }

    mountinfo->dirSize = args.dirSize;

    vfs_setfsprivate(mp, mountinfo);
    vfs_getnewfsid(mp);

    copyinstr(args.mntfromname, vfs_statfs(mp)->f_mntfromname, MNAMELEN - 1, &size);
    memset(vfs_statfs(mp)->f_mntfromname + size, 0, MNAMELEN - size);

    if (! createRootVnode(mp, mountinfo)) {
      goto bad;
    }

    vfs_setauthopaque(mp);

    return (0);

  bad:

    --URDFS::Kernel::mountCount;

    if (mountinfo) {
      if (mountinfo->socketName) {
        FREE(mountinfo->socketName, M_TEMP);
      }
      FREE(mountinfo, M_TEMP);

      vfs_setfsprivate(mp, NULL);
    }

    return error;
  }

  // --------------------------------------------------
  int
  urdfs_start(struct mount *mp, int flags, vfs_context_t context)
  {
#pragma unused(mp, flags, context)
    START_MARKER(__PRETTY_FUNCTION__);
    return 0;
  }


  int
  urdfs_unmount(struct mount *mp, int mntflags, vfs_context_t context)
  {
#pragma unused(context)
    START_MARKER(__PRETTY_FUNCTION__);

    vnode_t rootvp = URDFS::Kernel::mount2mountinfo(mp)->rootVnode;
    URDFS::Kernel::MountInfo *mountinfo = URDFS::Kernel::mount2mountinfo(mp);

    int flags = 0;
    if (mntflags & MNT_FORCE) {
      flags |= FORCECLOSE;
    }

    // flush all except root vnode
    int error = vflush(mp, rootvp, flags);
    if (error) return RETURN(error, __PRETTY_FUNCTION__);

    // see if anyone is using the root vnode besides the reference we took in urdfs_mount
    if (vnode_isinuse(rootvp, 1) && !(flags & FORCECLOSE)) {
      return RETURN(EBUSY, __PRETTY_FUNCTION__);
    }

    URDFS::Bridge::UnMount::Request request;
    URDFS::Kernel::Bridge::sendmsg(URDFS::Bridge::URDFS_UNMOUNT, mountinfo,
                                   &request, sizeof(request),
                                   NULL, 0);

    vnode_rele(rootvp);
    error = vflush(mp, NULLVP, flags);
    vfs_setfsprivate(mp, NULL);

    FREE(mountinfo->socketName, M_TEMP);
    FREE(mountinfo, M_TEMP);

    --URDFS::Kernel::mountCount;

    return RETURN(error, __PRETTY_FUNCTION__);
  }

  int
  urdfs_root(struct mount *mp, struct vnode **vpp, vfs_context_t context)
  {
#pragma unused(context)
    START_MARKER(__PRETTY_FUNCTION__);

    vnode_t vp = URDFS::Kernel::mount2mountinfo(mp)->rootVnode;
    int error = vnode_get(vp);
    if (error) {
      *vpp = NULLVP;
    } else {
      *vpp = vp;
    }

    return RETURN(error, __PRETTY_FUNCTION__);
  }


  int
  urdfs_vfs_getattr(struct mount *mp, struct vfs_attr *sbp, vfs_context_t context)
  {
#pragma unused(context)
    START_MARKER(__PRETTY_FUNCTION__);

    URDFS::Bridge::StatFs::Reply reply;
    memset(&reply, 0, sizeof(reply));

    URDFS::Kernel::MountInfo *mountinfo = URDFS::Kernel::mount2mountinfo(mp);
    URDFS::Bridge::StatFs::Request request;
    int error = URDFS::Kernel::Bridge::sendmsg(URDFS::Bridge::URDFS_STATFS, mountinfo,
                                           &request, sizeof(request),
                                           &reply, sizeof(reply));

    VFSATTR_RETURN(sbp, f_bsize, reply.fs_attr.f_bsize);
    VFSATTR_RETURN(sbp, f_iosize, reply.fs_attr.f_iosize);

    VFSATTR_RETURN(sbp, f_blocks, reply.fs_attr.f_blocks);
    VFSATTR_RETURN(sbp, f_bfree, reply.fs_attr.f_bfree);
    VFSATTR_RETURN(sbp, f_bavail, reply.fs_attr.f_bfree);

    VFSATTR_RETURN(sbp, f_files, reply.fs_attr.f_files);
    VFSATTR_RETURN(sbp, f_ffree, reply.fs_attr.f_ffree);

    return RETURN(error, __PRETTY_FUNCTION__);
  }

  // ------------------------------------------------------------
  struct vfsops urdfs_vfsops = {
    urdfs_mount,
    urdfs_start,
    urdfs_unmount,
    urdfs_root,
    NULL,
    urdfs_vfs_getattr,
    (int (*)(struct mount *, int, vfs_context_t))(nullop), // sync
    NULL,
    NULL,
    NULL,
    urdfs_init,
    NULL,
    NULL,
    {0}
  };
}

// ------------------------------------------------------------
kern_return_t
urdfs_module_start(struct kmod_info *ki, void *data)
{
#pragma unused(ki, data)
  START_MARKER(__PRETTY_FUNCTION__);

  errno_t error;
  struct vfs_fsentry vfe;

  memset(&vfe, 0, sizeof(vfe));
  vfe.vfe_vfsops = &urdfs_vfsops;
  vfe.vfe_vopcnt = 1;
  vfe.vfe_opvdescs = URDFS::Kernel::vnodeop_opv_desc_list;
  vfe.vfe_fstypenum = 0;
  strncpy(vfe.vfe_fsname, URDFS::Bridge::fsname, strlen(URDFS::Bridge::fsname));
  vfe.vfe_flags = VFS_TBLNOTYPENUM | VFS_TBL64BITREADY;

  error = vfs_fsadd(&vfe, &URDFS::Kernel::vfsconf);

  return (error ? KERN_FAILURE : KERN_SUCCESS);
}

// ------------------------------------------------------------
kern_return_t
urdfs_module_stop(struct kmod_info *ki, void *data)
{
#pragma unused(ki, data)
  START_MARKER(__PRETTY_FUNCTION__);

  int error;

  if (URDFS::Kernel::mountCount == 0) {
    error = vfs_fsremove(URDFS::Kernel::vfsconf);
    if (error == 0) {
      URDFS::NodeHash::terminate();
    }
  } else {
    error = EBUSY;
  }

  return (error ? KERN_FAILURE : KERN_SUCCESS);
}
