#include "urdfs_kext.hpp"
extern "C" {
#include <sys/ubc.h>
#include <sys/fcntl.h>
#include <sys/dirent.h>
#include <sys/unistd.h>
}
#include "urdfs_kext_node.hpp"
#include "urdfs_kext_bridge.hpp"

namespace {
  enum {
    STATUS_DIRTY =              1 << 0,
    STATUS_DELETED =            1 << 1,
    STATUS_READDIR_NOT_LOADED = 1 << 2,
  };

  void
  copyNameFromComponentname(char *name, const componentname *cnp)
  {
    strncpy(name, cnp->cn_nameptr, cnp->cn_namelen);
    name[cnp->cn_namelen] = '\0';
  }
}

// ======================================================================
/*
  struct vnop_lookup_args {
  struct vnodeop_desc *a_desc;
  vnode_t a_dvp;
  vnode_t *a_vpp;
  struct componentname *a_cnp;
  vfs_context_t a_context;
  };
*/
static int
urdfs_vnop_lookup(struct vnop_lookup_args *ap)
{
  START_MARKER(__PRETTY_FUNCTION__);

  vnode_t *vpp = ap->a_vpp;
  *vpp = NULLVP;

  componentname *cnp = ap->a_cnp;
  int nameiop = cnp->cn_nameiop;
  bool islastcn = cnp->cn_flags & ISLASTCN;

  // ----------------------------------------
  bool isdotdot = false;
  bool isdot = false;
  if (cnp->cn_flags & ISDOTDOT) {
    isdotdot = true;
  } else if (cnp->cn_nameptr[0] == '.' && cnp->cn_namelen == 1) {
    isdot = true;
  }

  // ----------------------------------------
  const vnode_t dvp = ap->a_dvp;

  if (cnp->cn_namelen > NAME_MAX) {
    return RETURN(ENAMETOOLONG, __PRETTY_FUNCTION__);

  } else if (! vnode_isdir(dvp)) {
    return RETURN(ENOTDIR, __PRETTY_FUNCTION__);

  } else if (isdotdot && (vnode_isvroot(dvp))) {
    printf("urdfs_vnop_lookup: invalid '..' from root\n");
    return RETURN(EIO, __PRETTY_FUNCTION__);

  } else if (islastcn && vnode_vfsisrdonly(dvp) && (nameiop == DELETE || nameiop == RENAME)) {
    return RETURN(EROFS, __PRETTY_FUNCTION__);
  }

  // ----------------------------------------
  int error = cache_lookup(dvp, vpp, cnp);
  switch (error) {
  case -1: // hit from cache
    return RETURN(0, __PRETTY_FUNCTION__);

  case 0: // no match or aged out
    if (isdot) {
      if ((nameiop == DELETE) && islastcn) {
        error = vnode_get(dvp);
        if (! error) {
          *vpp = dvp;
        }
        return RETURN(error, __PRETTY_FUNCTION__);
      }

      if ((nameiop == RENAME) && islastcn) {
        return RETURN(EISDIR, __PRETTY_FUNCTION__);
      }

      error = vnode_get(dvp);
      if (! error) {
        *vpp = dvp;
      }
      return RETURN(error, __PRETTY_FUNCTION__);
    }

    if (isdotdot) {
      vnode_t vp = URDFS::vnode2fsnode(dvp)->parent;
      error = vnode_get(vp);
      if (! error) {
        *vpp = vp;
      }
      return RETURN(error, __PRETTY_FUNCTION__);
    }

    URDFS::Bridge::Lookup::Reply reply;
    memset(&reply, 0, sizeof(reply));

    URDFS::Bridge::Lookup::Request request;
    request.parentIno = URDFS::vnode2fsnode(dvp)->ino;
    copyNameFromComponentname(request.name, cnp);

    URDFS::Kernel::MountInfo *mountinfo = URDFS::Kernel::mount2mountinfo(vnode_mount(dvp));
    error = URDFS::Kernel::Bridge::sendmsg(URDFS::Bridge::URDFS_LOOKUP, mountinfo,
                                           &request, sizeof(request),
                                           &reply, sizeof(reply));
    if (error) {
      if (error == ERESTART) {
        return RETURN(error, __PRETTY_FUNCTION__);
      }

      if ((nameiop == CREATE || nameiop == RENAME) && islastcn) {
        return RETURN(EJUSTRETURN, __PRETTY_FUNCTION__);
      }

      return RETURN(ENOENT, __PRETTY_FUNCTION__);
    }
    error = URDFS::Kernel::getVnode(vnode_mount(dvp), dvp, 0, cnp, reply.ino,
                                    reply.vtype,
                                    reply.atime, reply.mtime, reply.ctime,
                                    reply.filesize, vpp);
    return RETURN(error, __PRETTY_FUNCTION__);

  case ENOENT:
    return RETURN(ENOENT, __PRETTY_FUNCTION__);

  default: // unexpected error
    printf("urdfs_vnop_lookup: unexpected response from cache_lookup: %d\n", error);
    return RETURN(error, __PRETTY_FUNCTION__);
  }
}

// ======================================================================
/*
  struct vnop_open_args {
  struct vnodeop_desc *a_desc;
  vnode_t a_vp;
  int a_mode;
  vfs_context_t a_context;
  };
*/
static int
urdfs_vnop_open(struct vnop_open_args *ap)
{
  START_MARKER(__PRETTY_FUNCTION__);

  vnode_t vp = ap->a_vp;

  if (! vnode_isreg(vp) && ! vnode_isdir(vp)) {
    return RETURN(EFTYPE, __PRETTY_FUNCTION__);
  }

  URDFS::Node *pt = URDFS::vnode2fsnode(vp);

  // from cache
  if (pt->cacheVnode) {
    ++pt->opencount;
    if (pt->opencount == 0) {
      --pt->opencount;
      return RETURN(ENFILE, __PRETTY_FUNCTION__);
    }
    if (vnode_vtype(vp) == VDIR) {
      pt->status |= STATUS_READDIR_NOT_LOADED;
    }
    return RETURN(0, __PRETTY_FUNCTION__);
  }

  URDFS::Bridge::Open::Request request;
  request.flags = OFLAGS(ap->a_mode);
  request.ino = pt->ino;

  URDFS::Bridge::Open::Reply reply;
  memset(&reply, 0, sizeof(reply));

  URDFS::Kernel::MountInfo *mountinfo = URDFS::Kernel::mount2mountinfo(vnode_mount(vp));
  int error = URDFS::Kernel::Bridge::sendmsg(URDFS::Bridge::URDFS_OPEN, mountinfo,
                                             &request, sizeof(request),
                                             &reply, sizeof(reply));
  if (error) return RETURN(error, __PRETTY_FUNCTION__);

  vnode_t cachevp;
  error = vnode_lookup(reply.cacheFilePath, 0, &cachevp, ap->a_context);
  if (error) return RETURN(EIO, __PRETTY_FUNCTION__);

  vnode_ref(cachevp);
  vnode_put(cachevp);
  pt->cacheVnode = cachevp;

  pt->opencount = 1;

  if (vnode_isdir(vp)) {
    pt->status |= STATUS_READDIR_NOT_LOADED;
    vnode_setnocache(vp);
  }

  return RETURN(0, __PRETTY_FUNCTION__);
}

// ======================================================================
static int
syncUBCsize(vnode_t vp, const vfs_context_t context)
{
  START_MARKER(__PRETTY_FUNCTION__);

  URDFS::Node *pt = URDFS::vnode2fsnode(vp);
  vnode_t cachevp = pt->cacheVnode;

  struct vnode_attr attrbuf;
  VATTR_INIT(&attrbuf);
  VATTR_WANTED(&attrbuf, va_flags);
  VATTR_WANTED(&attrbuf, va_data_size);
  int error = vnode_getattr(cachevp, &attrbuf, context);
  if (error) return RETURN(error, __PRETTY_FUNCTION__);

  if (pt->filesize != static_cast<off_t>(attrbuf.va_data_size)) {
    ubc_setsize(vp, attrbuf.va_data_size);
    pt->filesize = attrbuf.va_data_size;
  }

  return RETURN(0, __PRETTY_FUNCTION__);
}


// ======================================================================
/*
  struct vnop_fsync_args {
  struct vnodeop_desc *a_desc;
  vnode_t a_vp;
  int a_waitfor;
  vfs_context_t a_context;
  };
*/
static int
urdfs_fsync(struct vnop_fsync_args *ap)
{
  START_MARKER(__PRETTY_FUNCTION__);

  vnode_t vp = ap->a_vp;
  URDFS::Node *pt = URDFS::vnode2fsnode(vp);

  if (! (pt->status & STATUS_DIRTY) || (pt->status & STATUS_DELETED)) {
    return RETURN(0, __PRETTY_FUNCTION__);
  }

  syncUBCsize(vp, ap->a_context);

  pt->status &= ~STATUS_DIRTY;

  URDFS::Bridge::Fsync::Request request;
  request.ino = pt->ino;

  URDFS::Kernel::MountInfo *mountinfo = URDFS::Kernel::mount2mountinfo(vnode_mount(vp));
  int error = URDFS::Kernel::Bridge::sendmsg(URDFS::Bridge::URDFS_FSYNC, mountinfo,
                                             &request, sizeof(request),
                                             NULL, 0);
  return RETURN(error, __PRETTY_FUNCTION__);
}

// ======================================================================
/*
  struct vnop_fsync_args {
  struct vnodeop_desc *a_desc;
  vnode_t a_vp;
  int a_waitfor;
  vfs_context_t a_context;
  };
*/
static int
urdfs_vnop_fsync(struct vnop_fsync_args *ap)
{
  START_MARKER(__PRETTY_FUNCTION__);

  vnode_t vp = ap->a_vp;
  if (! vnode_isreg(vp)) {
    return RETURN(0, __PRETTY_FUNCTION__);
  }
  if (URDFS::vnode2fsnode(vp)->cacheVnode == NULLVP) {
    return RETURN(0, __PRETTY_FUNCTION__);
  }

  int error = urdfs_fsync(ap);
  return RETURN(error, __PRETTY_FUNCTION__);
}

// ======================================================================
static int
urdfs_close(vnode_t vp, vfs_context_t context, bool forceFsync)
{
  START_MARKER(__PRETTY_FUNCTION__);

  URDFS::Node *pt = URDFS::vnode2fsnode(vp);

  if (pt->cacheVnode == NULLVP) {
    return RETURN(0, __PRETTY_FUNCTION__);
  }

  // fsync if need.
  if (vnode_isreg(vp) && (forceFsync || (pt->opencount == 0))) {
    struct vnop_fsync_args fsync_args;

    fsync_args.a_vp = vp;
    fsync_args.a_waitfor = TRUE;
    fsync_args.a_context = context;

    int error = urdfs_fsync(&fsync_args);
    if (error) return RETURN(error, __PRETTY_FUNCTION__);
  }

  // close if need.
  if (pt->opencount == 0) {
    URDFS::Bridge::Close::Request request;
    request.ino = pt->ino;

    URDFS::Kernel::MountInfo *mountinfo = URDFS::Kernel::mount2mountinfo(vnode_mount(vp));
    int error = URDFS::Kernel::Bridge::sendmsg(URDFS::Bridge::URDFS_CLOSE, mountinfo,
                                               &request, sizeof(URDFS::Bridge::Close::Request),
                                               NULL, 0);
    if (error) return RETURN(error, __PRETTY_FUNCTION__);

    vnode_t temp = pt->cacheVnode;
    pt->cacheVnode = NULLVP;
    vnode_rele(temp);
  }

  return RETURN(0, __PRETTY_FUNCTION__);
}

// ======================================================================
/*
  struct vnop_close_args {
  struct vnodeop_desc *a_desc;
  vnode_t a_vp;
  int a_fflag;
  vfs_context_t a_context;
  };
*/
static int
urdfs_vnop_close(struct vnop_close_args *ap)
{
  START_MARKER(__PRETTY_FUNCTION__);

  vnode_t vp = ap->a_vp;
  URDFS::Node *pt = URDFS::vnode2fsnode(vp);

  bool forceFsync = (((ap->a_fflag & FWRITE) != 0) && vnode_isreg(ap->a_vp));

  --pt->opencount;
  int error = urdfs_close(ap->a_vp, ap->a_context, forceFsync);
  return RETURN(error, __PRETTY_FUNCTION__);
}

// ======================================================================
/*
  struct vnop_read_args {
  struct vnodeop_desc *a_desc;
  vnode_t a_vp;
  struct uio *a_uio;
  int a_ioflag;
  vfs_context_t a_context;
  };
*/
static int
urdfs_rdwr(struct vnop_read_args *ap)
{
  START_MARKER(__PRETTY_FUNCTION__);

  vnode_t vp = ap->a_vp;
  if (vnode_isdir(vp)) {
    return RETURN(EISDIR, __PRETTY_FUNCTION__);
  }

  URDFS::Node *pt = URDFS::vnode2fsnode(vp);
  vnode_t cachevp = pt->cacheVnode;

  if (cachevp == NULLVP) {
    return RETURN(EIO, __PRETTY_FUNCTION__);
  }

  int error = syncUBCsize(vp, ap->a_context);
  if (error) return RETURN(error, __PRETTY_FUNCTION__);

  uio_t in_uio = ap->a_uio;

  if (uio_rw(in_uio) == UIO_READ) {
    // reading
    error = VNOP_READ(cachevp, in_uio, ap->a_ioflag, ap->a_context);
    return RETURN(error, __PRETTY_FUNCTION__);

  } else {
    // writing
    error = VNOP_WRITE(cachevp, in_uio, ap->a_ioflag, ap->a_context);

    pt->status |= STATUS_DIRTY;

    // ----------------------------------------
    struct vnode_attr attrbuf;
    VATTR_INIT(&attrbuf);
    VATTR_WANTED(&attrbuf, va_data_size);
    error = vnode_getattr(cachevp, &attrbuf, ap->a_context);

    if (uio_offset(in_uio) > static_cast<off_t>(attrbuf.va_data_size)) {
      struct vnode_attr vattr;
      VATTR_INIT(&vattr);
      VATTR_SET(&vattr, va_data_size, uio_offset(in_uio));
      error = vnode_setattr(cachevp, &vattr, ap->a_context);
      if (error) return RETURN(error, __PRETTY_FUNCTION__);

      ubc_setsize(vp, uio_offset(in_uio));
    }
    return RETURN(error, __PRETTY_FUNCTION__);
  }
}

// ======================================================================
/*
  struct vnop_read_args {
  struct vnodeop_desc *a_desc;
  vnode_t a_vp;
  struct uio *a_uio;
  int a_ioflag;
  vfs_context_t a_context;
  };
*/
static int
urdfs_vnop_read(struct vnop_read_args *ap)
{
  START_MARKER(__PRETTY_FUNCTION__);

  int error = urdfs_rdwr(ap);

  return RETURN(error, __PRETTY_FUNCTION__);
}

// ======================================================================
/*
  struct vnop_write_args {
  struct vnodeop_desc *a_desc;
  vnode_t a_vp;
  struct uio *a_uio;
  int a_ioflag;
  vfs_context_t a_context;
  };
*/
static int urdfs_vnop_write(struct vnop_write_args *ap)
{
  START_MARKER(__PRETTY_FUNCTION__);

  int error = urdfs_rdwr((struct vnop_read_args *)ap);

  return RETURN(error, __PRETTY_FUNCTION__);
}

// ======================================================================
/*
  struct vnop_getattr_args {
  struct vnodeop_desc *a_desc;
  vnode_t a_vp;
  struct vnode_attr *a_vap;
  vfs_context_t a_context;
  };
*/
static int
urdfs_vnop_getattr(struct vnop_getattr_args *ap)
{
  START_MARKER(__PRETTY_FUNCTION__);

  vnode_t vp = ap->a_vp;
  URDFS::Node *pt = URDFS::vnode2fsnode(vp);

  struct vnode_attr *vap = ap->a_vap;
  VATTR_RETURN(vap, va_nlink, 1);
  VATTR_RETURN(vap, va_uid, URDFS::Kernel::UNKNOWNUID);
  VATTR_RETURN(vap, va_gid, URDFS::Kernel::UNKNOWNUID);
  VATTR_RETURN(vap, va_fsid, vfs_statfs(vnode_mount(vp))->f_fsid.val[0]);
  VATTR_RETURN(vap, va_fileid, pt->ino);
  VATTR_RETURN(vap, va_access_time, pt->atime);
  VATTR_RETURN(vap, va_modify_time, pt->mtime);
  VATTR_RETURN(vap, va_change_time, pt->ctime);
  VATTR_RETURN(vap, va_gen, 0);
  VATTR_RETURN(vap, va_flags, 0);
  VATTR_RETURN(vap, va_rdev, 0);
  VATTR_RETURN(vap, va_filerev, 0);

  URDFS::Bridge::GetAttr::Reply reply;
  memset(&reply, 0, sizeof(reply));

  URDFS::Bridge::GetAttr::Request request;
  request.ino = pt->ino;

  URDFS::Kernel::MountInfo *mountinfo = URDFS::Kernel::mount2mountinfo(vnode_mount(vp));
  int error = URDFS::Kernel::Bridge::sendmsg(URDFS::Bridge::URDFS_GETATTR, mountinfo,
                                             &request, sizeof(request),
                                             &reply, sizeof(reply));
  if (error) return RETURN(error, __PRETTY_FUNCTION__);

  VATTR_RETURN(vap, va_data_size, reply.attr.st_size);
  VATTR_RETURN(vap, va_total_alloc, reply.attr.st_blocks * S_BLKSIZE);
  VATTR_RETURN(vap, va_iosize, reply.attr.st_blksize);
  VATTR_RETURN(vap, va_mode, reply.attr.st_mode);

  VATTR_RETURN(vap, va_access_time, reply.attr.st_atimespec);
  VATTR_RETURN(vap, va_modify_time, reply.attr.st_mtimespec);
  VATTR_RETURN(vap, va_change_time, reply.attr.st_ctimespec);

  return RETURN(0, __PRETTY_FUNCTION__);
}

// ======================================================================
/*
  struct vnop_remove_args {
  struct vnodeop_desc *a_desc;
  vnode_t a_dvp;
  vnode_t a_vp;
  struct componentname *a_cnp;
  int a_flags;
  vfs_context_t a_context;
  };
*/
static int
urdfs_vnop_remove(struct vnop_remove_args *ap)
{
  START_MARKER(__PRETTY_FUNCTION__);

  vnode_t vp = ap->a_vp;
  if (vnode_isdir(vp)) {
    return RETURN(EPERM, __PRETTY_FUNCTION__);
  }
  if ((ap->a_flags & VNODE_REMOVE_NODELETEBUSY) && vnode_isinuse(vp, 0)) {
    return RETURN(EBUSY, __PRETTY_FUNCTION__);
  }
  cache_purge(vp);

  URDFS::Node *pt = URDFS::vnode2fsnode(vp);

  URDFS::Bridge::Remove::Request request;
  request.ino = pt->ino;

  URDFS::Kernel::MountInfo *mountinfo = URDFS::Kernel::mount2mountinfo(vnode_mount(vp));
  int error = URDFS::Kernel::Bridge::sendmsg(URDFS::Bridge::URDFS_REMOVE, mountinfo,
                                             &request, sizeof(request),
                                             NULL, 0);
  if (error) return RETURN(error, __PRETTY_FUNCTION__);

  pt->status |= STATUS_DELETED;
  vnode_recycle(vp);

  return RETURN(0, __PRETTY_FUNCTION__);
}

// ======================================================================
/*
  struct vnop_rmdir_args {
  struct vnodeop_desc *a_desc;
  vnode_t a_dvp;
  vnode_t a_vp;
  struct componentname *a_cnp;
  vfs_context_t a_context;
  };
*/
static
int urdfs_vnop_rmdir(struct vnop_rmdir_args *ap)
{
  START_MARKER(__PRETTY_FUNCTION__);

  vnode_t vp = ap->a_vp;
  cache_purge(vp);

  URDFS::Node *pt = URDFS::vnode2fsnode(vp);

  URDFS::Bridge::RmDir::Request request;
  request.ino = pt->ino;

  URDFS::Kernel::MountInfo *mountinfo = URDFS::Kernel::mount2mountinfo(vnode_mount(vp));
  int error = URDFS::Kernel::Bridge::sendmsg(URDFS::Bridge::URDFS_RMDIR, mountinfo,
                                             &request, sizeof(request),
                                             NULL, 0);
  if (error) return RETURN(error, __PRETTY_FUNCTION__);

  pt->status |= STATUS_DELETED;
  vnode_recycle(vp);

  return RETURN(0, __PRETTY_FUNCTION__);
}

// ======================================================================
/*
  struct vnop_create_args {
  struct vnodeop_desc *a_desc;
  vnode_t a_dvp;
  vnode_t *a_vpp;
  struct componentname *a_cnp;
  struct vnode_attr *a_vap;
  vfs_context_t a_context;
  };
*/
static int
urdfs_vnop_create(struct vnop_create_args *ap)
{
  START_MARKER(__PRETTY_FUNCTION__);

  vnode_t *vpp = ap->a_vpp;
  *vpp = NULLVP;

  if (ap->a_vap->va_type != VREG) {
    return RETURN(ENOTSUP, __PRETTY_FUNCTION__);
  }

  struct componentname *cnp = ap->a_cnp;
  vnode_t dvp = ap->a_dvp;
  URDFS::Node *dpt = URDFS::vnode2fsnode(dvp);

  URDFS::Bridge::Create::Request request;
  request.parentIno = dpt->ino;
  request.mode = ap->a_vap->va_mode;
  copyNameFromComponentname(request.name, cnp);

  URDFS::Bridge::Create::Reply reply;
  memset(&reply, 0, sizeof(reply));

  URDFS::Kernel::MountInfo *mountinfo = URDFS::Kernel::mount2mountinfo(vnode_mount(dvp));
  int error = URDFS::Kernel::Bridge::sendmsg(URDFS::Bridge::URDFS_CREATE, mountinfo,
                                             &request, sizeof(request),
                                             &reply, sizeof(reply));
  if (error) return RETURN(error, __PRETTY_FUNCTION__);

  dpt->status |= STATUS_READDIR_NOT_LOADED;

  // ----------------------------------------
  // set vpp
  struct timeval tv;
  struct timespec ts;
  microtime(&tv);
  TIMEVAL_TO_TIMESPEC(&tv, &ts);

  error = URDFS::Kernel::getVnode(vnode_mount(dvp), dvp, 0, cnp, reply.ino, VREG, ts, ts, ts, 0, vpp);
  return RETURN(error, __PRETTY_FUNCTION__);
}

// ======================================================================
/*
  struct vnop_rename_args {
  struct vnodeop_desc *a_desc;
  vnode_t a_fdvp;
  vnode_t a_fvp;
  struct componentname *a_fcnp;
  vnode_t a_tdvp;
  vnode_t a_tvp;
  struct componentname *a_tcnp;
  vfs_context_t a_context;
  };
*/
static int urdfs_vnop_rename(struct vnop_rename_args *ap)
{
  START_MARKER(__PRETTY_FUNCTION__);

  vnode_t fvp = ap->a_fvp;
  vnode_t tvp = ap->a_tvp;
  vnode_t tdvp = ap->a_tdvp;
  vnode_t fdvp = ap->a_fdvp;
  struct componentname *tcnp = ap->a_tcnp;

  cache_purge(fvp);

  URDFS::Bridge::Rename::Request request;
  request.from_parentIno = URDFS::vnode2fsnode(fdvp)->ino;
  request.from_ino = URDFS::vnode2fsnode(fvp)->ino;
  request.to_parentIno = URDFS::vnode2fsnode(tdvp)->ino;
  request.to_ino = (tvp != NULLVP) ? URDFS::vnode2fsnode(tvp)->ino : 0;
  copyNameFromComponentname(request.to_name, tcnp);

  URDFS::Kernel::MountInfo *mountinfo = URDFS::Kernel::mount2mountinfo(vnode_mount(fvp));
  int error = URDFS::Kernel::Bridge::sendmsg(URDFS::Bridge::URDFS_RENAME, mountinfo,
                                             &request, sizeof(request),
                                             NULL, 0);
  if (error) return RETURN(error, __PRETTY_FUNCTION__);

  if (tvp != NULLVP) {
    if (tvp != fvp) {
      cache_purge(tvp);
    }

    URDFS::vnode2fsnode(tvp)->status |= STATUS_DELETED;
    vnode_recycle(tvp);
  }

  URDFS::vnode2fsnode(fdvp)->status |= STATUS_READDIR_NOT_LOADED;
  URDFS::vnode2fsnode(tdvp)->status |= STATUS_READDIR_NOT_LOADED;

  return RETURN(0, __PRETTY_FUNCTION__);
}

// ======================================================================
/*
  struct vnop_mkdir_args {
  struct vnodeop_desc *a_desc;
  vnode_t a_dvp;
  vnode_t *a_vpp;
  struct componentname *a_cnp;
  struct vnode_attr *a_vap;
  vfs_context_t a_context;
  };
*/
static int
urdfs_vnop_mkdir(struct vnop_mkdir_args *ap)
{
  START_MARKER(__PRETTY_FUNCTION__);

  vnode_t *vpp = ap->a_vpp;
  *vpp = NULLVP;

  if (ap->a_vap->va_type != VDIR) {
    return RETURN(ENOTDIR, __PRETTY_FUNCTION__);
  }

  struct componentname *cnp = ap->a_cnp;
  vnode_t dvp = ap->a_dvp;
  URDFS::Node *dpt = URDFS::vnode2fsnode(dvp);

  URDFS::Bridge::MkDir::Request request;
  request.parentIno = dpt->ino;
  request.mode = ap->a_vap->va_mode;
  copyNameFromComponentname(request.name, cnp);

  URDFS::Bridge::MkDir::Reply reply;
  memset(&reply, 0, sizeof(reply));

  URDFS::Kernel::MountInfo *mountinfo = URDFS::Kernel::mount2mountinfo(vnode_mount(dvp));
  int error = URDFS::Kernel::Bridge::sendmsg(URDFS::Bridge::URDFS_MKDIR, mountinfo,
                                             &request, sizeof(request),
                                             &reply, sizeof(reply));
  if (error) return RETURN(error, __PRETTY_FUNCTION__);

  dpt->status |= STATUS_READDIR_NOT_LOADED;

  // ----------------------------------------
  // set vpp
  struct timeval tv;
  struct timespec ts;
  microtime(&tv);
  TIMEVAL_TO_TIMESPEC(&tv, &ts);

  error = URDFS::Kernel::getVnode(vnode_mount(dvp), dvp, 0, cnp, reply.ino, VDIR, ts, ts, ts, mountinfo->dirSize, vpp);
  return RETURN(error, __PRETTY_FUNCTION__);
}

// ======================================================================
/*
  struct vnop_setattr_args {
  struct vnodeop_desc *a_desc;
  vnode_t a_vp;
  struct vnode_attr *a_vap;
  vfs_context_t a_context;
  };
*/
static int
urdfs_vnop_setattr(struct vnop_setattr_args *ap)
{
  START_MARKER(__PRETTY_FUNCTION__);

  vnode_t vp = ap->a_vp;
  if (vnode_isvroot(vp)) {
    return RETURN(EACCES, __PRETTY_FUNCTION__);
  }

  URDFS::Node *pt = URDFS::vnode2fsnode(vp);
  struct vnode_attr *vap = ap->a_vap;

  // ----------------------------------------
  // change real
  URDFS::Bridge::SetAttr::Reply reply;
  memset(&reply, 0, sizeof(reply));

  URDFS::Bridge::SetAttr::Request request;
  memset(&request, 0, sizeof(request));
  request.ino = pt->ino;
  if (VATTR_IS_ACTIVE(vap, va_data_size)) {
    request.isSetSize = true;
    request.size = vap->va_data_size;
  }
  if (VATTR_IS_ACTIVE(vap, va_mode)) {
    request.isSetMode = true;
    request.mode = vap->va_mode;
  }
  if (VATTR_IS_ACTIVE(vap, va_access_time)) {
    request.isSetAtime = true;
    request.atime = vap->va_access_time;
  }
  if (VATTR_IS_ACTIVE(vap, va_modify_time)) {
    request.isSetMtime = true;
    request.mtime = vap->va_modify_time;
  }

  URDFS::Kernel::MountInfo *mountinfo = URDFS::Kernel::mount2mountinfo(vnode_mount(vp));
  int error = URDFS::Kernel::Bridge::sendmsg(URDFS::Bridge::URDFS_SETATTR, mountinfo,
                                             &request, sizeof(request),
                                             &reply, sizeof(reply));
  if (error) return RETURN(error, __PRETTY_FUNCTION__);

  // ----------------------------------------
  // change kernel-cache & UBC
  vnode_t cachevp = pt->cacheVnode;
  if (cachevp == NULLVP) {
    printf("%s cachevp == NULLVP", __PRETTY_FUNCTION__);
    return RETURN(0, __PRETTY_FUNCTION__);
  }

  if (VATTR_IS_ACTIVE(vap, va_data_size) && vnode_isreg(vp)) {
    struct vnode_attr attrbuf;
    VATTR_INIT(&attrbuf);
    VATTR_WANTED(&attrbuf, va_flags);
    VATTR_WANTED(&attrbuf, va_data_size);
    error = vnode_getattr(cachevp, &attrbuf, ap->a_context);
    if (error) return RETURN(error, __PRETTY_FUNCTION__);

    if (vap->va_data_size != attrbuf.va_data_size || static_cast<off_t>(vap->va_data_size) != pt->filesize) {
      pt->status |= STATUS_DIRTY;
    }

    ubc_setsize(vp, static_cast<off_t>(vap->va_data_size));
    pt->filesize = vap->va_data_size;
  }

  error = vnode_setattr(cachevp, vap, ap->a_context);

  return RETURN(0, __PRETTY_FUNCTION__);
}

// ======================================================================
/*
  struct vnop_readdir_args {
  struct vnodeop_desc *a_desc;
  vnode_t a_vp;
  struct uio *a_uio;
  int a_flags;
  int *a_eofflag;
  int *a_numdirent;
  vfs_context_t a_context;
  };
*/
static int
urdfs_vnop_readdir(struct vnop_readdir_args *ap)
{
  START_MARKER(__PRETTY_FUNCTION__);

  vnode_t vp = ap->a_vp;
  if (! vnode_isdir(vp)) {
    return RETURN(ENOTDIR, __PRETTY_FUNCTION__);
  }

  URDFS::Node *pt = URDFS::vnode2fsnode(vp);
  vnode_t cachevp = pt->cacheVnode;

  if (cachevp == NULLVP) {
    return RETURN(EINVAL, __PRETTY_FUNCTION__);
  }

  if (ap->a_flags & VNODE_READDIR_REQSEEKOFF) {
    return RETURN(EINVAL, __PRETTY_FUNCTION__);
  }

  uio_t uio = ap->a_uio;

  if (uio_offset(uio) == 0 || (pt->status & STATUS_READDIR_NOT_LOADED)) {
    URDFS::Bridge::ReadDir::Request request;
    request.ino = pt->ino;

    URDFS::Kernel::MountInfo *mountinfo = URDFS::Kernel::mount2mountinfo(vnode_mount(vp));
    int error = URDFS::Kernel::Bridge::sendmsg(URDFS::Bridge::URDFS_READDIR, mountinfo,
                                               &request, sizeof(request),
                                               NULL, 0);
    if (error) {
      pt->status |= STATUS_READDIR_NOT_LOADED;
      return RETURN(error, __PRETTY_FUNCTION__);
    }

    pt->status &= ~ STATUS_READDIR_NOT_LOADED;
  }

  if (ap->a_flags & VNODE_READDIR_EXTENDED) {
    return RETURN(EINVAL, __PRETTY_FUNCTION__);
  }

  if (((uio_offset(uio) % sizeof(struct dirent)) != 0) ||
      (uio_resid(uio) < static_cast<user_ssize_t>(sizeof(struct dirent)))) {
    return RETURN(EINVAL, __PRETTY_FUNCTION__);
  }

  user_ssize_t count = uio_resid(uio);
  count -= (uio_offset(uio) + count) % sizeof(struct dirent);
  if (count <= 0) {
    return RETURN(EINVAL, __PRETTY_FUNCTION__);
  }

  user_ssize_t lost = uio_resid(uio) - count;
  uio_setresid(uio, count);

  int error = VNOP_READ(cachevp, uio, 0, ap->a_context);
  uio_setresid(uio, uio_resid(uio) + lost);

  if (ap->a_eofflag) {
    struct vnode_attr vattr;
    VATTR_INIT(&vattr);
    VATTR_WANTED(&vattr, va_data_size);
    error = vnode_getattr(cachevp, &vattr, ap->a_context);
    if (error) return RETURN(error, __PRETTY_FUNCTION__);

    *ap->a_eofflag = static_cast<off_t>(vattr.va_data_size) <= uio_offset(uio);
  }

  return RETURN(0, __PRETTY_FUNCTION__);
}

// ======================================================================
/*
  struct vnop_readlink_args {
  struct vnodeop_desc *a_desc;
  vnode_t a_vp;
  struct uio *a_uio;
  vfs_context_t a_context;
  }
*/
static int
urdfs_vnop_readlink(struct vnop_readlink_args *ap)
{
  START_MARKER(__PRETTY_FUNCTION__);

  vnode_t vp = ap->a_vp;
  if (! vnode_islnk(vp)) {
    return RETURN(EINVAL, __PRETTY_FUNCTION__);
  }

  URDFS::Node *pt = URDFS::vnode2fsnode(vp);

  URDFS::Bridge::ReadLink::Request request;
  request.ino = pt->ino;

  URDFS::Bridge::ReadLink::Reply reply;
  memset(&reply, 0, sizeof(reply));

  URDFS::Kernel::MountInfo *mountinfo = URDFS::Kernel::mount2mountinfo(vnode_mount(vp));
  int error = URDFS::Kernel::Bridge::sendmsg(URDFS::Bridge::URDFS_READLINK, mountinfo,
                                             &request, sizeof(request),
                                             &reply, sizeof(reply));
  if (error) return RETURN(error, __PRETTY_FUNCTION__);

  error = uiomove(reply.path, strlen(reply.path), ap->a_uio);
  return RETURN(error, __PRETTY_FUNCTION__);
}

// ======================================================================
/*
  struct vnop_symlink_args {
  struct vnodeop_desc *a_desc;
  vnode_t a_dvp;
  vnode_t *a_vpp;
  struct componentname *a_cnp;
  struct vnode_attr *a_vap;
  char *a_target;
  vfs_context_t a_context;
  }
*/
static int
urdfs_vnop_symlink(struct vnop_symlink_args *ap)
{
  START_MARKER(__PRETTY_FUNCTION__);

  vnode_t *vpp = ap->a_vpp;
  *vpp = NULLVP;

  if (ap->a_vap->va_type != VLNK) {
    return RETURN(EINVAL, __PRETTY_FUNCTION__);
  }

  if (ap->a_target[0] == '\0') {
    return RETURN(EINVAL, __PRETTY_FUNCTION__);
  }

  struct componentname *cnp = ap->a_cnp;
  vnode_t dvp = ap->a_dvp;
  URDFS::Node *dpt = URDFS::vnode2fsnode(dvp);

  URDFS::Bridge::SymLink::Request request;
  request.parentIno = dpt->ino;
  copyNameFromComponentname(request.name, cnp);
  snprintf(request.linktarget, sizeof(request.linktarget), "%s", ap->a_target);

  URDFS::Bridge::SymLink::Reply reply;
  memset(&reply, 0, sizeof(reply));

  URDFS::Kernel::MountInfo *mountinfo = URDFS::Kernel::mount2mountinfo(vnode_mount(dvp));
  int error = URDFS::Kernel::Bridge::sendmsg(URDFS::Bridge::URDFS_SYMLINK, mountinfo,
                                             &request, sizeof(request),
                                             &reply, sizeof(reply));
  if (error) return RETURN(error, __PRETTY_FUNCTION__);

  dpt->status |= STATUS_READDIR_NOT_LOADED;

  // ----------------------------------------
  // set vpp
  struct timeval tv;
  struct timespec ts;
  microtime(&tv);
  TIMEVAL_TO_TIMESPEC(&tv, &ts);

  error = URDFS::Kernel::getVnode(vnode_mount(dvp), dvp, 0, cnp, reply.ino, VLNK, ts, ts, ts, 0, vpp);
  return RETURN(error, __PRETTY_FUNCTION__);
}

// ======================================================================
/*
  struct vnop_inactive_args {
  struct vnodeop_desc *a_desc;
  vnode_t a_vp;
  vfs_context_t a_context;
  };
*/
static int
urdfs_vnop_inactive(struct vnop_inactive_args *ap)
{
#pragma unused(ap)
  START_MARKER(__PRETTY_FUNCTION__);

  return RETURN(0, __PRETTY_FUNCTION__);
}

// ======================================================================
/*
  struct vnop_reclaim_args {
  struct vnodeop_desc *a_desc;
  vnode_t a_vp;
  vfs_context_t a_context;
  };
*/
static int
urdfs_vnop_reclaim(struct vnop_reclaim_args *ap)
{
  START_MARKER(__PRETTY_FUNCTION__);

  vnode_t vp = ap->a_vp;
  vnode_removefsref(vp);

  URDFS::Node *pt = URDFS::vnode2fsnode(vp);
  URDFS::NodeHash::remove(pt);

  cache_purge(vp);
  vnode_clearfsnode(vp);
  FREE(pt, M_TEMP);

  return RETURN(0, __PRETTY_FUNCTION__);
}

// ======================================================================
/*
  struct vnop_pathconf_args {
  struct vnodeop_desc *a_desc;
  vnode_t a_vp;
  int a_name;
  register_t *a_retval;
  vfs_context_t a_context;
  };
*/
static int
urdfs_vnop_pathconf(struct vnop_pathconf_args *ap)
{
  START_MARKER(__PRETTY_FUNCTION__);

  switch (ap->a_name) {
  case _PC_LINK_MAX:
    *ap->a_retval = 1;
    return RETURN(0, __PRETTY_FUNCTION__);

  case _PC_NAME_MAX:
    *ap->a_retval = NAME_MAX;
    return RETURN(0, __PRETTY_FUNCTION__);

  case _PC_PATH_MAX:
    *ap->a_retval = PATH_MAX;
    return RETURN(0, __PRETTY_FUNCTION__);

  case _PC_CHOWN_RESTRICTED:
    *ap->a_retval = 1;
    return RETURN(0, __PRETTY_FUNCTION__);

  case _PC_NO_TRUNC:
    *ap->a_retval = 1;
    return RETURN(0, __PRETTY_FUNCTION__);

  case _PC_PIPE_BUF:
  default:
    *ap->a_retval = -1;
    return RETURN(EINVAL, __PRETTY_FUNCTION__);
  }
}

// ======================================================================
typedef errno_t (*VNodeOp)(void *);
int( **urdfs_vnodeop_p)();

struct vnodeopv_entry_desc urdfs_vnodeop_entries[] = {
  {&vnop_default_desc, reinterpret_cast<VNodeOp>(vn_default_error)},
  {&vnop_lookup_desc, reinterpret_cast<VNodeOp>(urdfs_vnop_lookup)},
  {&vnop_create_desc, reinterpret_cast<VNodeOp>(urdfs_vnop_create)},
  {&vnop_open_desc, reinterpret_cast<VNodeOp>(urdfs_vnop_open)},
  {&vnop_close_desc, reinterpret_cast<VNodeOp>(urdfs_vnop_close)},
  {&vnop_getattr_desc, reinterpret_cast<VNodeOp>(urdfs_vnop_getattr)},
  {&vnop_setattr_desc, reinterpret_cast<VNodeOp>(urdfs_vnop_setattr)},
  {&vnop_read_desc, reinterpret_cast<VNodeOp>(urdfs_vnop_read)},
  {&vnop_write_desc, reinterpret_cast<VNodeOp>(urdfs_vnop_write)},
  {&vnop_fsync_desc, reinterpret_cast<VNodeOp>(urdfs_vnop_fsync)},
  {&vnop_remove_desc, reinterpret_cast<VNodeOp>(urdfs_vnop_remove)},
  {&vnop_rename_desc, reinterpret_cast<VNodeOp>(urdfs_vnop_rename)},
  {&vnop_mkdir_desc, reinterpret_cast<VNodeOp>(urdfs_vnop_mkdir)},
  {&vnop_rmdir_desc, reinterpret_cast<VNodeOp>(urdfs_vnop_rmdir)},
  {&vnop_readdir_desc, reinterpret_cast<VNodeOp>(urdfs_vnop_readdir)},
  {&vnop_readlink_desc, reinterpret_cast<VNodeOp>(urdfs_vnop_readlink)},
  {&vnop_symlink_desc, reinterpret_cast<VNodeOp>(urdfs_vnop_symlink)},
  {&vnop_inactive_desc, reinterpret_cast<VNodeOp>(urdfs_vnop_inactive)},
  {&vnop_reclaim_desc, reinterpret_cast<VNodeOp>(urdfs_vnop_reclaim)},
  {&vnop_pathconf_desc, reinterpret_cast<VNodeOp>(urdfs_vnop_pathconf)},
  { NULL, NULL },
};

namespace URDFS {
  namespace Kernel {
    struct vnodeopv_desc vnodeop_opv_desc = {
      reinterpret_cast<int (***)(void*)>(&urdfs_vnodeop_p),
      urdfs_vnodeop_entries,
    };
  }
}
