#include <sys/mount.h>
#include <sys/dirent.h>

#include "sshfs_op.hpp"
#include "urdfs.hpp"
#include "urdfs_utils.hpp"

bool
SSHFS::Operator::initialize(const std::string &sshhostpath, uid_t _uid, gid_t _gid)
{
  START_MARKER(__PRETTY_FUNCTION__);

  uid = _uid;
  gid = _gid;

  if (! sftp.initialize(sshhostpath)) {
    return false;
  }
  return true;
}

void
SSHFS::Operator::setuidgid(const std::string &path)
{
  URDFS::Utils::chown(path, uid, gid);
}

void
SSHFS::Operator::eraseStatCache(const std::string &path)
{
  mutex.lock();
  {
    statcache.erase(path);
  }
  mutex.unlock();
}

int
SSHFS::Operator::create(const std::string &path)
{
  START_MARKER(__PRETTY_FUNCTION__);

  URDFS::TmpFile tmpfile; // upload zero-size file.
  int error = sftp.upload(tmpfile.getFileName(), path, true);
  if (error) RETURN(error, __PRETTY_FUNCTION__);

  eraseStatCache(path);

  return RETURN(0, __PRETTY_FUNCTION__);
}

int
SSHFS::Operator::open(const URDFS::NodeEntry &node, bool writeAccess, bool trunc)
{
#pragma unused(writeAccess)

  START_MARKER(__PRETTY_FUNCTION__);

  std::string remote;
  node.getPath(&remote);

  eraseStatCache(remote);

  struct stat st;
  int error = getattr(&st, remote);
  if (error) return RETURN(EIO, __PRETTY_FUNCTION__);

  if (S_ISREG(st.st_mode)) {
    if (trunc) {
      truncate(node.getCacheFileName().c_str(), 0);
    } else {
      // check mtime
      struct stat cachestat;
      error = URDFS::Utils::stat(node.getCacheFileName(), &cachestat);
      if (error == 0) {
        // cachefile exist
        if (cachestat.st_mtime == st.st_mtime && cachestat.st_size == st.st_size) {
          // from cache
          return RETURN(0, __PRETTY_FUNCTION__);
        }
      }

      error = sftp.download(remote, node.getCacheFileName(), true);
      if (error) return RETURN(error, __PRETTY_FUNCTION__);
    }
  }

  setuidgid(node.getCacheFileName());

  return RETURN(0, __PRETTY_FUNCTION__);
}

int
SSHFS::Operator::getattr(struct stat *st, const std::string &path)
{
  START_MARKER(__PRETTY_FUNCTION__);

  bool cachehit = false;
  int error = 0;
  mutex.lock();
  {
    typeof(statcache.end()) p = statcache.find(path);
    time_t now = time(NULL);
    // from cache
    if (p != statcache.end() && p->second->lifetime > now) {
      cachehit = true;
      if (! p->second->exist) {
        error = ENOENT;
      } else {
        *st = p->second->st;
      }
    }
  }
  mutex.unlock();
  if (cachehit) return RETURN(error, __PRETTY_FUNCTION__);


  SFTP::Attrib attr;
  error = sftp.lstat(&attr, path);
  if (error) {
    if (error == ENOENT) {
      mutex.lock();
      {
        StatCachePtr p = StatCachePtr(new StatCache);
        p->lifetime = time(NULL) + STATCACHE_LIFETIME;
        p->exist = false;
        statcache[path] = p;
      }
      mutex.unlock();
    }
    return RETURN(error, __PRETTY_FUNCTION__);
  }

  st->st_size = attr.size;
  st->st_uid = attr.uid;
  st->st_gid = attr.gid;
  st->st_mode = attr.perm;
  st->st_atime = attr.atime;
  st->st_mtime = attr.mtime;

  st->st_dev = 0;
  st->st_nlink = 1;
  st->st_rdev = 0;
  st->st_blksize = SSHFS::IOSIZE;
  st->st_flags = 0;
  st->st_gen = 0;
  st->st_blocks = (st->st_size + S_BLKSIZE - 1) / S_BLKSIZE;

  mutex.lock();
  {
    StatCachePtr p = StatCachePtr(new StatCache);
    p->lifetime = time(NULL) + STATCACHE_LIFETIME;
    p->exist = true;
    p->st = *st;
    statcache[path] = p;
  }
  mutex.unlock();

  return RETURN(0, __PRETTY_FUNCTION__);
}

int
SSHFS::Operator::setattr(const struct stat &st, const std::string &path)
{
  START_MARKER(__PRETTY_FUNCTION__);

  SFTP::Attrib attr;
  attr.perm = st.st_mode;
  attr.size = st.st_size;
  attr.atime = st.st_mode;
  attr.mtime = st.st_mode;

  int error = sftp.setattr(path, attr);
  if (error) RETURN(error, __PRETTY_FUNCTION__);

  eraseStatCache(path);

  return RETURN(0, __PRETTY_FUNCTION__);
}

int
SSHFS::Operator::fsync(const std::string &path, const std::string &cacheFilePath)
{
  START_MARKER(__PRETTY_FUNCTION__);

  int error = sftp.upload(cacheFilePath, path, true);
  if (error) RETURN(error, __PRETTY_FUNCTION__);

  eraseStatCache(path);

  return RETURN(0, __PRETTY_FUNCTION__);
}

int
SSHFS::Operator::mkdir(const std::string &path)
{
  START_MARKER(__PRETTY_FUNCTION__);

  int error = sftp.mkdir(path);
  if (error) return RETURN(error, __PRETTY_FUNCTION__);

  eraseStatCache(path);

  return RETURN(0, __PRETTY_FUNCTION__);
}

int
SSHFS::Operator::rmdir(const std::string &path)
{
  START_MARKER(__PRETTY_FUNCTION__);

  int error = sftp.rmdir(path);
  if (error) return RETURN(error, __PRETTY_FUNCTION__);

  eraseStatCache(path);

  return RETURN(0, __PRETTY_FUNCTION__);
}

int
SSHFS::Operator::rename(const std::string &newpath, const std::string &oldpath)
{
  START_MARKER(__PRETTY_FUNCTION__);

  int error = sftp.rename(oldpath, newpath);
  if (error) return RETURN(error, __PRETTY_FUNCTION__);

  eraseStatCache(oldpath);
  eraseStatCache(newpath);

  return RETURN(0, __PRETTY_FUNCTION__);
}

int
SSHFS::Operator::remove(const std::string &path)
{
  START_MARKER(__PRETTY_FUNCTION__);

  int error = sftp.rm(path);
  if (error) return RETURN(error, __PRETTY_FUNCTION__);

  eraseStatCache(path);

  return RETURN(0, __PRETTY_FUNCTION__);
}

int
SSHFS::Operator::statfs(struct vfsstatfs *vsf)
{
  START_MARKER(__PRETTY_FUNCTION__);

  memset(vsf, 0, sizeof(struct vfsstatfs));

  vsf->f_bsize = SSHFS::BLOCKSIZE;
  vsf->f_iosize = SSHFS::IOSIZE;

  vsf->f_blocks = SSHFS::NUMBLOCKS;
  vsf->f_bfree = SSHFS::NUMBLOCKS;
  vsf->f_bavail = SSHFS::NUMBLOCKS;

  vsf->f_files = SSHFS::NUMFILES;
  vsf->f_ffree = SSHFS::NUMFILES - 2;

  return RETURN(0, __PRETTY_FUNCTION__);
}

int
SSHFS::Operator::readdir(struct dirent **ent, int *entnum, URDFS::NodeEntry *node)
{
  std::string remote;
  node->getPath(&remote);

  std::vector<SFTP::DirentPtr> dirent;
  int error = sftp.readdir(&dirent, remote);
  if (error) return RETURN(EIO, __PRETTY_FUNCTION__);

  *entnum = dirent.size();

  if (ent) {
    *ent = new struct dirent[*entnum];

    struct dirent *ep = *ent;
    for (typeof(dirent.begin()) p = dirent.begin(); p != dirent.end(); ++p, ++ep) {
      std::string &filename = (*p)->filename;
      SFTP::Attrib &attrib = (*p)->a;

      memset(ep, 0, sizeof(struct dirent));
      ep->d_reclen = sizeof(struct dirent);
      ep->d_type = URDFS::Utils::mode2dtype(attrib.perm);
      ep->d_namlen = filename.size();
      snprintf(ep->d_name, sizeof(ep->d_name), filename.c_str());

      if (filename == ".") {
        ep->d_ino = node->getIno();

      } else if (filename == "..") {
        ep->d_ino = node->getParent()->getIno();

      } else {
        URDFS::NodeEntry *tmp = node->findChild(filename);
        if (tmp == NULL) {
          tmp = node->makeChild(filename);
        }
        ep->d_ino = tmp->getIno();
      }
    }
  }

  return RETURN(0, __PRETTY_FUNCTION__);
}

int
SSHFS::Operator::readlink(std::string *referred, const std::string &path)
{
  START_MARKER(__PRETTY_FUNCTION__);

  int error = sftp.readlink(referred, path);
  return RETURN(error, __PRETTY_FUNCTION__);
}

int
SSHFS::Operator::symlink(const std::string &path, const std::string &target)
{
  START_MARKER(__PRETTY_FUNCTION__);

  int error = sftp.symlink(target, path);
  return RETURN(error, __PRETTY_FUNCTION__);
}
