#include "urdfs.hpp"
#include "urdfs_mount.hpp"
#include "urdfs_server.hpp"

#include <sys/socket.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <paths.h>
#include <string>

extern std::string kextpath;

namespace {
  void
  parentExit(int x)
  {
#pragma unused(x)
    exit(EXIT_SUCCESS);
  }
}

bool
URDFS::Mount::isAlreadyMounted(const std::string &mntfromname)
{
  START_MARKER(__PRETTY_FUNCTION__);

  if (mntfromname.empty()) {
    return false;
  }

  struct statfs *buffer;
  int count = getmntinfo(&buffer, MNT_NOWAIT);

  for (int i = 0; i < count; i++) {
    struct statfs *p = buffer + i;

    if (p->f_owner == getuid() &&
        strcmp(URDFS::Bridge::fsname, p->f_fstypename) == 0 &&
        mntfromname == p->f_mntfromname) {
      return true;
    }
  }

  return false;
}

bool
URDFS::Mount::loadKext(void)
{
  std::cerr << "Loading Kernel Extension" << std::endl;

  pid_t pid = fork();
  if (pid < 0) {
    return false;
  }

  if (pid == 0) {
    const char *cmd = "/sbin/kextload";
    execl(cmd, cmd, "-t", kextpath.c_str(), NULL);
  }

  union wait status;
  if ((wait4(pid, (int *)&status, 0, NULL) == pid) && (WIFEXITED(status))) {
    return true;
  } else {
    return false;
  }
}

namespace {
  void
  makeTmpDir(const std::string &mountpoint)
  {
    struct stat st;
    stat(mountpoint.c_str(), &st);

    URDFS::TmpFile::initializeTmpBaseDir();
    chown(URDFS::TmpFile::getTmpBaseDir().c_str(), st.st_uid, st.st_gid);
  }
}

int
URDFS::Mount::mount(URDFS::Operator *op, const std::string &mountpoint, const std::string &mntfromname)
{
  std::string volumename = mountpoint.substr(mountpoint.rfind('/') + 1);

  if (isAlreadyMounted(mntfromname)) {
    std::cerr << mountpoint << " is already mounted" << std::endl;
    return EBUSY;
  }

  signal(SIGTERM, parentExit);

  pid_t pid = fork();
  if (pid < 0) return errno;

  if (pid > 0) {
    // parent process
    union wait status;
    if ((wait4(pid, (int *)&status, 0, NULL) == pid) && (WIFEXITED(status))) {
      return status.w_retcode;
    } else {
      return ECHILD;
    }
  }

  // ----------------------------------------
  chdir("/");
  signal(SIGPIPE, SIG_IGN);

  // ----------------------------------------
  struct vfsconf vfc;
  vfc.vfc_typenum = -1;
  int error = getvfsbyname(URDFS::Bridge::fsname, &vfc);
  if (error) {
    if (! loadKext()) {
      return false;
    }
    getvfsbyname(URDFS::Bridge::fsname, &vfc);
  }

  // ----------------------------------------
  makeTmpDir(mountpoint);
  URDFS::NodeEntry::makeRootNode();

  URDFS::Server server(op);

  // ----------------------------------------
  URDFS::Bridge::MountArgs args;
  args.mntfromname = const_cast<char *>(mntfromname.c_str());
  args.version = 1;
  const struct sockaddr_un &un = server.getSockAddr();
  args.socketNameLen = sizeof(un);
  args.socketName = (struct sockaddr *)&un;
  args.volName = const_cast<char *>(volumename.c_str());
  args.flags = 0;

  // ----------------------------------------
  int mntflags = 0;
  ::mount(URDFS::Bridge::fsname, mountpoint.c_str(), mntflags, &args);

  // ----------------------------------------
  kill(getppid(), SIGTERM);
  server.doLoop();

  return 0;
}
