#include "urdfs.hpp"
#include "urdfs_node.hpp"

ino_t URDFS::NodeEntry::nextIno = 0;
std::map<ino_t, URDFS::NodeEntryPtr> URDFS::NodeEntry::mapIno2Node;
URDFS::Thread::Mutex URDFS::NodeEntry::mutex;

URDFS::NodeEntry::NodeEntry(URDFS::NodeEntry *_parent, const std::string &_name)
{
  START_MARKER(__PRETTY_FUNCTION__);

  ino = nextIno++;
  name = _name;
  if (_parent == NULL) {
    // root node
    parent = this;
  } else {
    parent = _parent;
  }

  mapIno2Node[ino] = URDFS::NodeEntryPtr(this);
}

URDFS::NodeEntry *
URDFS::NodeEntry::makeRootNode(void)
{
  START_MARKER(__PRETTY_FUNCTION__);

  nextIno = URDFS::Bridge::ROOT_INO;
  URDFS::NodeEntry *p = new URDFS::NodeEntry(NULL, ".");
  return p;
}

void
URDFS::NodeEntry::internal_deleteSelf(void)
{
  START_MARKER(__PRETTY_FUNCTION__);

  typeof(children.begin()) it = children.begin();
  for (;;) {
    if (it == children.end()) break;

    URDFS::NodeEntry *p = it->second;
    ++it;

    p->internal_deleteSelf();
  }

  (parent->children).erase(name);
  mapIno2Node.erase(ino);
}

void
URDFS::NodeEntry::deleteSelf(void)
{
  START_MARKER(__PRETTY_FUNCTION__);

  mutex.lock();
  internal_deleteSelf();
  mutex.unlock();
}

URDFS::NodeEntry *
URDFS::NodeEntry::findChild(const std::string &_name)
{
  START_MARKER(__PRETTY_FUNCTION__);

  typeof(children.end()) p = children.find(_name);
  if (p == children.end()) {
    return NULL;
  } else {
    return p->second;
  }
}

URDFS::NodeEntry *
URDFS::NodeEntry::makeChild(const std::string &_name)
{
  START_MARKER(__PRETTY_FUNCTION__);

  mutex.lock();

  URDFS::NodeEntry *p = findChild(_name);
  if (p == NULL) {
    URDFS::NodeEntry *c = new URDFS::NodeEntry(this, _name);
    children[_name] = c;
    p = c;
  }

  mutex.unlock();

  return p;
}

void
URDFS::NodeEntry::getPath(std::string *path) const
{
  START_MARKER(__PRETTY_FUNCTION__);

  if (ino == URDFS::Bridge::ROOT_INO) {
    *path = ".";
  } else {
    parent->getPath(path);
    *path += '/' + name;
  }
}

URDFS::NodeEntry *
URDFS::NodeEntry::get(ino_t _ino)
{
  START_MARKER(__PRETTY_FUNCTION__);

  typeof(mapIno2Node.end()) p = mapIno2Node.find(_ino);
  if (p == mapIno2Node.end()) {
    return NULL;
  } else {
    return (p->second).get();
  }
}

void
URDFS::NodeEntry::rename(URDFS::NodeEntry *newParent, const std::string &newName)
{
  START_MARKER(__PRETTY_FUNCTION__);

  if (newName == name && newParent == parent) {
    return;
  }

  (parent->children).erase(name);
  name = newName;

  if (newParent == parent) {
    (parent->children)[name] = this;
  }

  if (newParent != parent) {
    URDFS::NodeEntry *existNode = newParent->findChild(name);
    if (existNode) {
      existNode->deleteSelf();
    }
    (newParent->children)[name] = this;
    parent = newParent;
  }
}
