#include <sys/select.h>
#include <sys/dirent.h>
#include <fcntl.h>
#include <paths.h>
#include <iostream>
#include <fstream>
#include <boost/shared_array.hpp>
#include <boost/lexical_cast.hpp>

#include "urdfs.hpp"
#include "urdfs_server.hpp"
#include "urdfs_utils.hpp"
#include "urdfs_mount.hpp"

URDFS::Server::Server(URDFS::Operator *_op)
{
  START_MARKER(__PRETTY_FUNCTION__);

  op = _op;

  // ----------------------------------------
  // setup thread pool
  pthread_attr_t attr;
  pthread_t thread;
  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
  for (int i = 0; i < NUM_THREAD; ++i) {
    pthread_create(&thread, &attr, handleRequestThread, this);
  }

  // ----------------------------------------
  makeSocket();
}

URDFS::Server::~Server(void)
{
  START_MARKER(__PRETTY_FUNCTION__);

  deleteSocket();
}

void
URDFS::Server::sendReply(int sock, void *data, size_t size, int error)
{
  struct iovec iov[2];
  iov[0].iov_base = reinterpret_cast<caddr_t>(&error);
  iov[0].iov_len = sizeof(error);
  if (size > 0) {
    iov[1].iov_base = reinterpret_cast<caddr_t>(data);
    iov[1].iov_len = size;
  }

  struct msghdr msg;
  memset(&msg, 0, sizeof(msg));
  msg.msg_iov = iov;
  if (size > 0) {
    msg.msg_iovlen = 2;
  } else {
    msg.msg_iovlen = 1;
  }

  sendmsg(sock, &msg, 0);
}

void
URDFS::Server::dispatchOperator(int sock)
{
  START_MARKER(__PRETTY_FUNCTION__);

  int operation;
  if (read(sock, &operation, sizeof(operation)) < 0) {
    sendReply(sock, NULL, 0, EINVAL);
    return;
  }
  size_t requestsize;
  if (read(sock, &requestsize, sizeof(requestsize)) < 0) {
    sendReply(sock, NULL, 0, EINVAL);
    return;
  }

  boost::shared_array<char> msg = boost::shared_array<char>(new char[requestsize]);
  if (read(sock, msg.get(), requestsize) < static_cast<int>(requestsize)) {
    sendReply(sock, NULL, 0, EINVAL);
    return;
  }

  URDFS::Bridge::Request *request = reinterpret_cast<URDFS::Bridge::Request *>(msg.get());

  // ------------------------------------------------------------
  switch (operation) {
  case URDFS::Bridge::URDFS_LOOKUP:
    {
      URDFS::Bridge::Lookup::Reply reply;
      int error = do_lookup(&reply, request->lookup);
      sendReply(sock, &reply, sizeof(reply), error);
      break;
    }
  case URDFS::Bridge::URDFS_CREATE:
    {
      URDFS::Bridge::Create::Reply reply;
      int error = do_create(&reply, request->create);
      sendReply(sock, &reply, sizeof(reply), error);
      break;
    }
  case URDFS::Bridge::URDFS_OPEN:
    {
      URDFS::Bridge::Open::Reply reply;
      int error = do_open(&reply, request->open);
      sendReply(sock, &reply, sizeof(reply), error);
      break;
    }
  case URDFS::Bridge::URDFS_CLOSE:
    {
      int error = do_close(request->close);
      sendReply(sock, NULL, 0, error);
      break;
    }
  case URDFS::Bridge::URDFS_GETATTR:
    {
      URDFS::Bridge::GetAttr::Reply reply;
      int error = do_getattr(&reply, request->getattr);
      sendReply(sock, &reply, sizeof(reply), error);
      break;
    }
  case URDFS::Bridge::URDFS_SETATTR:
    {
      int error = do_setattr(request->setattr);
      sendReply(sock, NULL, 0, error);
      break;
    }
  case URDFS::Bridge::URDFS_MKDIR:
    {
      URDFS::Bridge::MkDir::Reply reply;
      int error = do_mkdir(&reply, request->mkdir);
      sendReply(sock, &reply, sizeof(reply), error);
      break;
    }
  case URDFS::Bridge::URDFS_FSYNC:
    {
      int error = do_fsync(request->fsync);
      sendReply(sock, NULL, 0, error);
      break;
    }
  case URDFS::Bridge::URDFS_REMOVE:
    {
      int error = do_remove(request->remove);
      sendReply(sock, NULL, 0, error);
      break;
    }
  case URDFS::Bridge::URDFS_RENAME:
    {
      int error = do_rename(request->rename);
      sendReply(sock, NULL, 0, error);
      break;
    }
  case URDFS::Bridge::URDFS_RMDIR:
    {
      int error = do_rmdir(request->rmdir);
      sendReply(sock, NULL, 0, error);
      break;
    }
  case URDFS::Bridge::URDFS_READDIR:
    {
      int error = do_readdir(request->readdir);
      sendReply(sock, NULL, 0, error);
      break;
    }
  case URDFS::Bridge::URDFS_READLINK:
    {
      URDFS::Bridge::ReadLink::Reply reply;
      int error = do_readlink(&reply, request->readlink);
      sendReply(sock, &reply, sizeof(reply), error);
      break;
    }
  case URDFS::Bridge::URDFS_SYMLINK:
    {
      URDFS::Bridge::SymLink::Reply reply;
      int error = do_symlink(&reply, request->symlink);
      sendReply(sock, &reply, sizeof(reply), error);
      break;
    }
  case URDFS::Bridge::URDFS_STATFS:
    {
      URDFS::Bridge::StatFs::Reply reply;
      int error = do_statfs(&reply, request->statfs);
      sendReply(sock, &reply, sizeof(reply), error);
      break;
    }
  case URDFS::Bridge::URDFS_UNMOUNT:
    {
      sendReply(sock, NULL, 0, 0);
      exitLoopFlag = true;
      break;
    }
  default:
    {
      int error = ENOTSUP;
      sendReply(sock, NULL, 0, error);
      break;
    }
  }
}

void *
URDFS::Server::handleRequestThread(void *arg)
{
  START_MARKER(__PRETTY_FUNCTION__);

  URDFS::Server *serv = reinterpret_cast<URDFS::Server *>(arg);

#pragma unused(arg)
  START_MARKER(__PRETTY_FUNCTION__);

  URDFS::Thread::Mutex &mutex = (serv->requestsMutex);
  URDFS::Thread::Cond &cond = (serv->requestCond);
  std::queue<int> &queue = (serv->queueRequestSocket);

  for (;;) {
    mutex.lock();
    if (queue.empty()) {
      cond.wait(&mutex);
      mutex.unlock();
      continue;

    } else {
      int sock = queue.front();
      queue.pop();

      mutex.unlock();
      (serv->dispatchOperator)(sock);
      close(sock);
    }
  }
}

bool
URDFS::Server::enqueueRequest(int sock)
{
  START_MARKER(__PRETTY_FUNCTION__);

  requestsMutex.lock();

  queueRequestSocket.push(sock);
  requestCond.signal();

  requestsMutex.unlock();
  return true;
}

void
URDFS::Server::doLoop(void)
{
  START_MARKER(__PRETTY_FUNCTION__);

  listen(listenSocket, 128);

  exitLoopFlag = false;
  while (! exitLoopFlag) {
    fd_set readfds;
    FD_ZERO(&readfds);
    FD_SET(listenSocket, &readfds);

    int nfds = listenSocket + 1;
    struct timeval timeout;
    timeout.tv_sec = 1;
    timeout.tv_usec = 0;
    if (select(nfds, &readfds, NULL, NULL, &timeout) == -1) {
      return;
    }

    if (FD_ISSET(listenSocket, &readfds)) {
      struct sockaddr_un addr;
      socklen_t addrlen;

      int s = accept(listenSocket, reinterpret_cast<struct sockaddr *>(&addr), &addrlen);
      if (s < 0) {
        return;
      }
      enqueueRequest(s);
    }
  }
}

bool
URDFS::Server::makeSocket(void)
{
  START_MARKER(__PRETTY_FUNCTION__);

  listenSocket = socket(PF_LOCAL, SOCK_STREAM, 0);
  if (listenSocket < 0) {
    return false;
  }

  listenSocketPath = _PATH_TMP ".urdfs.socket.";
  listenSocketPath += boost::lexical_cast<std::string>(getpid());

  memset(&listenSocketAddr, 0, sizeof(listenSocketAddr));
  listenSocketAddr.sun_len = sizeof(listenSocketAddr);
  listenSocketAddr.sun_family = AF_LOCAL;
  snprintf(listenSocketAddr.sun_path, sizeof(listenSocketAddr.sun_path), "%s", listenSocketPath.c_str());

  // set socket writable for everyone.
  mode_t prevMask = umask(0555);
  if (bind(listenSocket, reinterpret_cast<struct sockaddr *>(&listenSocketAddr), sizeof(listenSocketAddr)) < 0) {
    return false;
  }
  umask(prevMask);

  // prevent to unlink socket
  chflags(listenSocketPath.c_str(), UF_IMMUTABLE);

  return true;
}

void
URDFS::Server::deleteSocket(void)
{
  START_MARKER(__PRETTY_FUNCTION__);

  if (! listenSocketPath.empty()) {
    const char *path = listenSocketPath.c_str();
    chflags(path, 0);
    unlink(path);
  }
}


// ======================================================================
int
URDFS::Server::do_lookup(URDFS::Bridge::Lookup::Reply *reply, const URDFS::Bridge::Lookup::Request &request)
{
  START_MARKER(__PRETTY_FUNCTION__);

  NodeEntry *parentNode = NodeEntry::get(request.parentIno);
  if (! parentNode) return EIO;

  std::string remote;
  parentNode->getPath(&remote);
  remote = remote + "/" + request.name;

  const NodeEntry *node = parentNode->findChild(request.name);
  if (node == NULL) {
    node = parentNode->makeChild(request.name);
  }

  struct stat statbuf;
  int error = op->getattr(&statbuf, remote);
  if (error) return error;

  reply->ino = node->getIno();
  reply->vtype = URDFS::Utils::mode2vtype(statbuf.st_mode);
  reply->atime.tv_sec = statbuf.st_atime;
  reply->mtime.tv_sec = statbuf.st_mtime;
  reply->ctime.tv_sec = statbuf.st_ctime;
  reply->filesize = statbuf.st_size;

  return RETURN(0, __PRETTY_FUNCTION__);
}

int
URDFS::Server::do_create(URDFS::Bridge::Create::Reply *reply, const URDFS::Bridge::Create::Request &request)
{
  START_MARKER(__PRETTY_FUNCTION__);

  NodeEntry *parentNode = NodeEntry::get(request.parentIno);
  if (! parentNode) return RETURN(EIO, __PRETTY_FUNCTION__);

  std::string remote;
  parentNode->getPath(&remote);
  remote = remote + "/" + request.name;

  int error = op->create(remote);
  if (error) return RETURN(error, __PRETTY_FUNCTION__);

  const NodeEntry *node = parentNode->findChild(request.name);
  if (node == NULL) {
    node = parentNode->makeChild(request.name);
  }

  reply->ino = node->getIno();

  return RETURN(0, __PRETTY_FUNCTION__);
}

int
URDFS::Server::do_open(URDFS::Bridge::Open::Reply *reply, const URDFS::Bridge::Open::Request &request)
{
  START_MARKER(__PRETTY_FUNCTION__);

  NodeEntry *node = NodeEntry::get(request.ino);
  if (! node) return RETURN(EIO, __PRETTY_FUNCTION__);

  std::string remote;
  node->getPath(&remote);

  struct stat st;
  int error = op->getattr(&st, remote);
  if (error) return RETURN(error, __PRETTY_FUNCTION__);

  bool writeAccess = ((request.flags & O_ACCMODE) != O_RDONLY);
  bool trunc = writeAccess && (request.flags & O_TRUNC);
  error = op->open(*node, writeAccess, trunc);
  if (error) return RETURN(EIO, __PRETTY_FUNCTION__);

  snprintf(reply->cacheFilePath, sizeof(reply->cacheFilePath), "%s", node->getCacheFileName().c_str());
  return RETURN(0, __PRETTY_FUNCTION__);
}

int
URDFS::Server::do_close(const URDFS::Bridge::Close::Request &request)
{
#pragma unused(request)
  START_MARKER(__PRETTY_FUNCTION__);

  return RETURN(0, __PRETTY_FUNCTION__);
}

int
URDFS::Server::do_getattr(URDFS::Bridge::GetAttr::Reply *reply, const URDFS::Bridge::GetAttr::Request &request)
{
  START_MARKER(__PRETTY_FUNCTION__);

  const NodeEntry *node = NodeEntry::get(request.ino);
  if (! node) return RETURN(EIO, __PRETTY_FUNCTION__);

  std::string remote;
  node->getPath(&remote);

  int error = op->getattr(&(reply->attr), remote);
  return RETURN(error, __PRETTY_FUNCTION__);
}

int
URDFS::Server::do_setattr(const URDFS::Bridge::SetAttr::Request &request)
{
  START_MARKER(__PRETTY_FUNCTION__);

  const NodeEntry *node = NodeEntry::get(request.ino);
  if (! node) return RETURN(EIO, __PRETTY_FUNCTION__);

  std::string remote;
  node->getPath(&remote);

  struct stat st;
  int error = op->getattr(&st, remote);
  if (error) return RETURN(error, __PRETTY_FUNCTION__);

  if (request.isSetSize) {
    st.st_size = request.size;
  }
  if (request.isSetMode) {
    st.st_mode = request.mode;
  }
  if (request.isSetAtime) {
    st.st_atimespec = request.atime;
  }
  if (request.isSetMtime) {
    st.st_mtimespec = request.mtime;
  }

  error = op->setattr(st, remote);

  return RETURN(error, __PRETTY_FUNCTION__);
}

int
URDFS::Server::do_fsync(const URDFS::Bridge::Fsync::Request &request)
{
  START_MARKER(__PRETTY_FUNCTION__);

  const NodeEntry *node = NodeEntry::get(request.ino);
  if (! node) return RETURN(EIO, __PRETTY_FUNCTION__);

  syncCacheFileMode(*node);

  std::string remote;
  node->getPath(&remote);

  int error = op->fsync(remote, node->getCacheFileName());
  return RETURN(error, __PRETTY_FUNCTION__);
}

int
URDFS::Server::do_remove(const URDFS::Bridge::Remove::Request &request)
{
  START_MARKER(__PRETTY_FUNCTION__);

  NodeEntry *node = NodeEntry::get(request.ino);
  if (! node) return EIO;

  std::string remote;
  node->getPath(&remote);

  int error = op->remove(remote);
  if (error) return error;

  node->deleteSelf();

  return 0;
}

int
URDFS::Server::do_rename(const URDFS::Bridge::Rename::Request &request)
{
  START_MARKER(__PRETTY_FUNCTION__);

  NodeEntry *oldNode = NodeEntry::get(request.from_ino);
  if (! oldNode) return EIO;

  NodeEntry *newParent = NodeEntry::get(request.to_parentIno);
  if (! newParent) return EIO;

  NodeEntry *newNode = NodeEntry::get(request.to_ino); // can be NULL.
  std::string name = request.to_name;

  std::string remoteOld;
  oldNode->getPath(&remoteOld);

  std::string remoteNew;
  newParent->getPath(&remoteNew);
  remoteNew = remoteNew + "/" + name;

  int error = op->rename(remoteNew, remoteOld);
  if (error) return error;

  if (newNode) {
    oldNode->deleteSelf();
  } else {
    oldNode->rename(newParent, name);
  }

  return 0;
}

int
URDFS::Server::do_mkdir(URDFS::Bridge::MkDir::Reply *reply, const URDFS::Bridge::MkDir::Request &request)
{
  START_MARKER(__PRETTY_FUNCTION__);

  NodeEntry *parentNode = NodeEntry::get(request.parentIno);
  if (! parentNode) return EIO;

  std::string remote;
  parentNode->getPath(&remote);
  remote = remote + "/" + request.name;

  int error = op->mkdir(remote);
  if (error) return error;

  const NodeEntry *node = parentNode->findChild(request.name);
  if (! node) {
    node = parentNode->makeChild(request.name);
  }
  reply->ino = node->getIno();

  return 0;
}

int
URDFS::Server::do_rmdir(const URDFS::Bridge::RmDir::Request &request)
{
  START_MARKER(__PRETTY_FUNCTION__);

  NodeEntry *node = NodeEntry::get(request.ino);
  if (! node) return EIO;

  // check directory empty
  int entnum;
  int error = op->readdir(NULL, &entnum, node);
  if (entnum > 2) {
    return ENOTEMPTY;
  }

  std::string remote;
  node->getPath(&remote);

  error = op->rmdir(remote);
  if (error) return error;

  node->deleteSelf();

  return 0;
}

int
URDFS::Server::do_readdir(const URDFS::Bridge::ReadDir::Request &request)
{
  START_MARKER(__PRETTY_FUNCTION__);

  NodeEntry *node = NodeEntry::get(request.ino);
  if (! node) return EIO;

  struct dirent *ent;
  int entnum;
  int error = op->readdir(&ent, &entnum, node);
  if (error) return EIO;

  std::ofstream file(node->getCacheFileName().c_str());
  file.write(reinterpret_cast<char *>(ent), sizeof(struct dirent) * entnum);

  delete[] ent;

  return 0;
}

int
URDFS::Server::do_readlink(URDFS::Bridge::ReadLink::Reply *reply, const URDFS::Bridge::ReadLink::Request &request)
{
  START_MARKER(__PRETTY_FUNCTION__);

  NodeEntry *node = NodeEntry::get(request.ino);
  if (! node) return EIO;

  std::string remote;
  node->getPath(&remote);

  std::string referred;
  int error = op->readlink(&referred, remote);
  if (error) return EIO;

  snprintf(reply->path, sizeof(reply->path), referred.c_str());
  return 0;
}

int
URDFS::Server::do_symlink(URDFS::Bridge::SymLink::Reply *reply, const URDFS::Bridge::SymLink::Request &request)
{
  START_MARKER(__PRETTY_FUNCTION__);

  NodeEntry *parentNode = NodeEntry::get(request.parentIno);
  if (! parentNode) return EIO;

  std::string remote;
  parentNode->getPath(&remote);
  remote = remote + "/" + request.name;

  int error = op->symlink(remote, request.linktarget);
  if (error) return EIO;

  const NodeEntry *node = parentNode->findChild(request.name);
  if (node == NULL) {
    node = parentNode->makeChild(request.name);
  }

  reply->ino = node->getIno();

  return 0;
}

int
URDFS::Server::do_statfs(URDFS::Bridge::StatFs::Reply *reply, const URDFS::Bridge::StatFs::Request &request)
{
#pragma unused(request)
  START_MARKER(__PRETTY_FUNCTION__);

  return op->statfs(&(reply->fs_attr));
}

int
URDFS::Server::syncCacheFileMode(const URDFS::NodeEntry &node)
{
  START_MARKER(__PRETTY_FUNCTION__);

  struct stat cachestat;
  int error = URDFS::Utils::stat(node.getCacheFileName(), &cachestat);
  if (error) RETURN(error, __PRETTY_FUNCTION__);

  std::string path;
  node.getPath(&path);

  struct stat st;
  error = op->getattr(&st, path);
  if (! error) {
    if (st.st_mode != cachestat.st_mode) {
      URDFS::Utils::chmod(node.getCacheFileName(), st.st_mode);
    }
  }

  return RETURN(0, __PRETTY_FUNCTION__);
}
